import datetime
import json
from six import string_types

from django.http import HttpResponse, HttpResponseRedirect
from django.core.urlresolvers import reverse, reverse_lazy, resolve
from django.core import serializers
from django.contrib import messages
from django.forms import Form
from django.forms.forms import NON_FIELD_ERRORS
from django.template import loader

from django.utils.html import strip_tags
from django.utils.translation import ugettext as _, ungettext

from django.db.models import Count, Avg
from django.shortcuts import get_object_or_404

from instances.views import InstanceFormMixin, InstanceViewMixin

from speeches.aggregates import Length
from speeches.forms import (
    SpeechForm, SpeechAudioForm, SectionForm,
    RecordingAPIForm, SpeakerForm, SectionPickForm,
    RecordingTimestampFormSet, SpeakerDeleteForm,
    PopoloImportForm, AkomaNtosoImportForm,
    )
from speeches.models import Speech, Speaker, Section, Recording, Tag
from speeches.mixins import Base32SingleObjectMixin, UnmatchingSlugException
from speeches.importers.import_akomantoso import ImportAkomaNtoso

from django.views.generic import (
    View, CreateView, UpdateView, DeleteView, DetailView, ListView,
    RedirectView, FormView,
    )
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import BaseFormView, BaseUpdateView

from django_select2.views import AutoResponseView

import logging

logger = logging.getLogger(__name__)


class AddAnSRedirectView(RedirectView):
    url = '/%(path)s%(suffix)s'
    permanent = True
    query_string = True
    suffix = 's'

    def get_redirect_url(self, **kwargs):
        kwargs['suffix'] = self.suffix
        return super(AddAnSRedirectView, self).get_redirect_url(**kwargs)


class JSONResponseMixin(object):
    """Mixin for returning HTTPResponse of JSON data"""

    def render_to_response(self, context, **kwargs):
        kwargs['content_type'] = 'application/json'
        if not isinstance(context, string_types):
            context = json.dumps(context)
        location = kwargs.pop('location', None)
        response = HttpResponse(context, **kwargs)
        if location:
            response['Location'] = location
        return response


class NamespaceMixin(object):
    """Mixin for adding current_app based on namespace"""

    def reverse(self, template, **kwargs):
        kwargs['current_app'] = resolve(self.request.path).namespace
        return reverse(template, **kwargs)

    def reverse_lazy(self, template, **kwargs):
        kwargs['current_app'] = resolve(self.request.path).namespace
        return reverse_lazy(template, **kwargs)

    def render_to_response(self, context, **kwargs):
        kwargs['current_app'] = resolve(self.request.path).namespace
        return super(NamespaceMixin, self).render_to_response(context, **kwargs)


class SpeechAudioCreate(JSONResponseMixin, BaseFormView):
    form_class = SpeechAudioForm
    http_method_names = ['post']

    def form_valid(self, form):
        # The cleaned_data contains the TemporaryUploadedFile
        # (File/UploadedFile subclass). The form.instance contains the
        # FieldFile with the magic save() we want.
        audio = form.cleaned_data['audio']
        form.instance.audio.save(audio.name, audio, save=False)
        return self.render_to_response(
            {'status': 'done', 'filename': form.instance.audio.name})

    def form_invalid(self, form):
        return self.render_to_response({'error': form.errors['audio']})


class SpeechMixin(NamespaceMixin, InstanceFormMixin):
    model = Speech
    form_class = SpeechForm

    def get_form(self, form_class=None):
        form = super(SpeechMixin, self).get_form(form_class)

        form.fields['section'].queryset = Section.objects.for_instance(self.request.instance)
        form.fields['speaker'].queryset = Speaker.objects.for_instance(self.request.instance)
        form.fields['tags'].queryset = Tag.objects.for_instance(self.request.instance)

        # Instance in the sense of https://github.com/mysociety/django-subdomain-instances
        form.fields['speaker'].instance = self.request.instance
        form.fields['section'].instance = self.request.instance

        return form


class SpeechDelete(SpeechMixin, DeleteView):
    def get_success_url(self):
        return self.reverse('speeches:parentless-list')


class SpeechCreate(SpeechMixin, CreateView):
    def get_context_data(self, **kwargs):
        context = super(SpeechCreate, self).get_context_data(**kwargs)
        added = self.request.GET.get('added', None)
        section = self.request.GET.get('section', None)
        if added and section:
            section_object = Section.objects.get(pk=section)
            if section_object:
                context['added'] = added
                context['section'] = section_object
        return context

    def get_initial(self):
        initial = super(SpeechCreate, self).get_initial()
        initial = initial.copy()
        try:
            speaker = int(self.request.GET['speaker'])
        except:
            speaker = None
        if speaker:
            try:
                initial['speaker'] = Speaker.objects.get(pk=speaker)
            except Speaker.DoesNotExist:
                # Ignore the supplied speaker
                pass
        try:
            section = int(self.request.GET['section'])
        except:
            section = None
        if section:
            try:
                section_object = Section.objects.get(pk=section)
                initial['section'] = section_object

                # default things to last speech in section
                try:
                    speech = section_object.speech_set.reverse()[0]

                    # Basic data is just defaulted from the last speech
                    initial['location'] = speech.location
                    initial['heading'] = speech.heading
                    initial['event'] = speech.event
                    initial['public'] = speech.public
                    initial['source_url'] = speech.source_url

                    # For speaker, to make it simpler to transcribe a dialogue
                    # between A -> B -> A -> B we default to the speaker of the
                    # *penultimate* speech, if that exists, or latest one
                    # otherwise.
                    if not speaker:
                        try:
                            penultimate_speech = section_object.speech_set.reverse()[1]
                            initial['speaker'] = penultimate_speech.speaker
                        except IndexError:
                            initial['speaker'] = speech.speaker

                    # We attempt to do vaguely clever things with the end or
                    # start time of previous speech.
                    if speech.end_date:
                        speech_end_datetime = datetime.datetime.combine(
                            speech.end_date, speech.end_time or datetime.time(0, 0))
                        if speech.end_time:
                            if speech.start_date and speech.start_time:
                                speech_start_datetime = datetime.datetime.combine(
                                    speech.start_date, speech.start_time)
                                if speech_start_datetime == speech_end_datetime:
                                    speech_end_datetime = speech_end_datetime + datetime.timedelta(seconds=1)
                            initial['start_time'] = speech_end_datetime.time()
                        initial['start_date'] = speech_end_datetime.date()
                    elif speech.start_date:
                        if speech.start_time:
                            speech_start_datetime = datetime.datetime.combine(speech.start_date, speech.start_time)
                            speech_start_datetime = speech_start_datetime + datetime.timedelta(seconds=10)
                            initial['start_time'] = speech_start_datetime.time()
                            initial['start_date'] = speech_start_datetime.date()
                        else:
                            initial['start_date'] = speech.start_date

                except IndexError:
                    # don't attempt to default anything
                    pass

            except Section.DoesNotExist:
                # Ignore the supplied section
                pass
        return initial

    def get_success_url(self):
        return "{}?created".format(self.object.get_absolute_url())

    def form_valid(self, form):
        resp = super(SpeechCreate, self).form_valid(form)

        # Now set off a Celery task to transcribe the audio for this speech
        self.object.start_transcribing()

        if 'add_another' in self.request.POST:
            url = self.reverse('speeches:speech-add')
            speech = self.object
            if speech.section_id:
                url = url + '?section=%d&added=%d' % (speech.section_id, speech.id)
            return HttpResponseRedirect(url)
        else:
            return resp


class SpeechUpdate(SpeechMixin, UpdateView):
    pass


class SpeechView(NamespaceMixin, InstanceViewMixin, DetailView):
    model = Speech

    def get_queryset(self):
        return super(SpeechView, self).get_queryset().visible(self.request)

    def get_context_data(self, **kwargs):
        context = super(SpeechView, self).get_context_data(**kwargs)
        context['title'] = strip_tags(
            self.object.title or
            u'\u201C%s\u201D' % self.object.summary
            )
        return context


class InstanceView(NamespaceMixin, InstanceViewMixin, ListView):
    """Done as a ListView on Speech to get recent speeches, we get instance for
    free in the request."""
    model = Speech
    paginate_by = 20

    # Use a slightly different template
    def get_template_names(self):
        return [
            "speeches/%s/home.html" % self.request.instance.label,
            "speeches/home.html"
        ]

    def get_context_data(self, **kwargs):
        context = super(InstanceView, self).get_context_data(**kwargs)
        context['count_speeches'] = Speech.objects.for_instance(self.request.instance).visible(self.request).count()
        context['count_sections'] = Section.objects.for_instance(self.request.instance).count()
        context['count_speakers'] = Speaker.objects.for_instance(self.request.instance).count()
        context['average_length'] = Speech.objects.for_instance(self.request.instance) \
            .annotate(length=Length('text')).aggregate(avg=Avg('length'))['avg']
        return context


# It doesn't actually use base32 IDs in the URL, but this works around Django
# 1.4 generic view bug, and allows non-canonical slug redirects to Just Work.
class SpeakerView(NamespaceMixin, InstanceViewMixin, Base32SingleObjectMixin, ListView):
    model = Speaker
    paginate_by = 50
    template_name = 'speeches/speaker_detail.html'
    slug_field = 'slugs__slug'

    def get_queryset(self):
        queryset = super(SpeakerView, self).get_queryset()
        self.object = self.get_object(queryset)
        return self.object.speech_set.all().visible(self.request) \
            .select_related('section', 'speaker').prefetch_related('tags').reverse()

    def get_context_data(self, **kwargs):
        kwargs['speech_list'] = self.object_list
        context = super(SpeakerView, self).get_context_data(**kwargs)
        context['section_count'] = self.object.speech_set.all().visible(self.request) \
            .aggregate(Count('section', distinct=True))['section__count']
        context['longest_speech'] = self.object.speech_set.annotate(length=Length('text')).order_by('-length')[:1]
        context['title'] = _('View Speaker: %(speaker_name)s') % {'speaker_name': self.object.name}
        return context


class SpeakerMixin(NamespaceMixin, InstanceFormMixin):
    model = Speaker
    form_class = SpeakerForm


class SpeakerList(NamespaceMixin, InstanceViewMixin, ListView):
    model = Speaker
    context_object_name = 'speaker_list'


class SpeakerCreate(SpeakerMixin, CreateView):
    pass


class SpeakerUpdate(SpeakerMixin, UpdateView):
    pass


class SpeakerDeleteNoSpeechesMixin(object):
    def delete(self):
        self.object.delete()
        return HttpResponseRedirect(self.get_success_url())

    def post(self, request, *args, **kwargs):
        if self.object.speech_set.exists():
            return super(SpeakerDeleteNoSpeechesMixin, self).post(request, *args, **kwargs)
        else:
            return self.delete()


class SpeakerDelete(SpeakerMixin, BaseUpdateView, SpeakerDeleteNoSpeechesMixin, FormView):
    form_class = SpeakerDeleteForm
    template_name = 'speeches/speaker_delete.html'

    def get_success_url(self):
        return self.reverse('speeches:speaker-list')

    def get_form(self, form_class=None):
        if self.object.speech_set.exists():
            form = super(SpeakerDelete, self).get_form(form_class)
            form.fields['new_speaker'].queryset = Speaker.objects.for_instance(self.request.instance)
            form.fields['new_speaker'].instance = self.request.instance
        else:
            form = Form

        return form

    def form_valid(self, form):
        action = form.cleaned_data['action']

        if action == 'Delete':
            Speech.objects.filter(speaker=self.object).delete()
        elif action == 'Narrative':
            # This, and reassignment below, will need to use a queue at scale
            for s in Speech.objects.filter(speaker=self.object):
                s.type = 'narrative'
                s.save()
        elif action == 'Reassign':
            for s in Speech.objects.filter(speaker=self.object):
                s.speaker = form.cleaned_data['new_speaker']
                s.save()

        return self.delete()


class ParentlessList(NamespaceMixin, InstanceViewMixin, ListView):
    model = Speech
    template_name = 'speeches/parentless_list.html'
    paginate_by = 50

    def get_queryset(self):
        qs = super(ParentlessList, self).get_queryset()
        qs = qs.visible(self.request)
        qs = qs.filter(section=None)
        qs = qs.select_related('speaker').prefetch_related('tags')
        return qs

    def get_context_data(self, **kwargs):
        context = super(ParentlessList, self).get_context_data(**kwargs)

        # Add in a QuerySet of all the sections without a parent
        context['section_list'] = Section.objects.for_instance(self.request.instance).filter(parent=None)
        return context


class SectionMixin(NamespaceMixin, InstanceFormMixin):
    model = Section
    form_class = SectionForm

    def get_form(self, form_class=None):
        form = super(SectionMixin, self).get_form(form_class)
        form.fields['parent'].queryset = Section.objects.for_instance(self.request.instance)
        form.fields['parent'].instance = self.request.instance
        return form


class SectionCreate(SectionMixin, CreateView):
    def get_initial(self):
        initial = super(SectionCreate, self).get_initial()
        initial = initial.copy()
        try:
            section = int(self.request.GET['section'])
        except:
            section = None
        if section:
            try:
                initial['parent'] = Section.objects.get(pk=section)
            except Section.DoesNotExist:
                # Ignore the supplied section
                pass
        return initial


class SectionUpdate(SectionMixin, UpdateView):
    pass


class SectionDelete(SectionMixin, DeleteView):
    def get_success_url(self):
        return self.reverse('speeches:parentless-list')


class SectionView(NamespaceMixin, InstanceViewMixin, DetailView):
    model = Section

    def get(self, request, *args, **kwargs):
        try:
            return super(SectionView, self).get(request, *args, **kwargs)
        except UnmatchingSlugException as e:
            return HttpResponseRedirect(e.args[0])

    def get_object(self, queryset=None):
        pk = self.kwargs.get(self.pk_url_kwarg, None)
        if pk is not None:
            return super(SectionView, self).get_object(queryset)
        full_slug = self.kwargs.get('full_slug', None)
        slugs = full_slug.split('/')
        parent = None
        for i, slug in enumerate(slugs):
            obj = get_object_or_404(self.model, instance=self.request.instance, slugs__slug=slug, parent=parent)
            if slug != obj.slug:
                new_url = obj.get_absolute_url()
                if i < len(slugs) - 1:
                    new_url += '/' + '/'.join(slugs[i + 1:])
                raise UnmatchingSlugException(new_url)
            parent = obj
        return obj

    def get_context_data(self, **kwargs):
        all_speeches = kwargs.pop('all_speeches', False)
        # Call the base implementation first to get a context
        context = super(SectionView, self).get_context_data(**kwargs)
        # Add in a QuerySet of all the speeches in this section
        context['section_tree'] = kwargs['object'].get_descendants_tree_with_speeches(
            self.request,
            all_speeches=all_speeches,
        )
        context['title'] = _('View Section: %(section_title)s') % {'section_title': self.object.title}
        context['speech_template'] = loader.get_template('speeches/speech.html')
        return context


class SectionViewAN(SectionView):
    template_name = 'speeches/section_detail.an'

    def render_to_response(self, context, **response_kwargs):
        response_kwargs['content_type'] = 'text/xml'
        return super(SectionView, self).render_to_response(context, **response_kwargs)

    def get_context_data(self, **kwargs):
        kwargs['all_speeches'] = True
        context = super(SectionViewAN, self).get_context_data(**kwargs)
        speakers = set(
            s[0].speaker
            for s in context['section_tree']
            if isinstance(s[0], Speech) and s[0].speaker
        )
        context['speakers'] = speakers
        context['server_name'] = self.request.META.get('SERVER_NAME')
        return context


class BothObjectAndFormMixin(object):
    def get_context_data(self, **kwargs):
        context = super(BothObjectAndFormMixin, self).get_context_data(**kwargs)
        if 'object' not in context:
            context['object'] = self.get_object()
        if 'form' not in context:
            context['form'] = SectionPickForm()
        # Restrict to request's instance (can we think of a nicer way to do this?)
        context['form'].fields['section'].queryset = Section.objects.for_instance(self.request.instance)
        return context


class RecordingList(NamespaceMixin, InstanceViewMixin, ListView):
    model = Recording


class RecordingDisplay(NamespaceMixin, BothObjectAndFormMixin, InstanceViewMixin, DetailView):
    model = Recording


class RecordingSetSection(NamespaceMixin, BothObjectAndFormMixin, InstanceFormMixin, FormView, SingleObjectMixin):
    template_name = 'speeches/recording_detail.html'
    form_class = SectionPickForm
    model = Recording

    def get_success_url(self):
        return self.object.get_absolute_url()

    def form_valid(self, form):
        self.object = self.get_object()
        self.object.add_speeches_to_section(form.cleaned_data['section'])
        messages.add_message(self.request, messages.SUCCESS, _("Speeches assigned."))
        return super(RecordingSetSection, self).form_valid(form)


class RecordingView(NamespaceMixin, View):
    def get(self, request, *args, **kwargs):
        view = RecordingDisplay.as_view()
        return view(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        view = RecordingSetSection.as_view()
        return view(request, *args, **kwargs)


# NB: this is not an UpdateView, at least for now, as it only updates
# RecordingTimestamp not the actual Recording class.
class RecordingUpdate(NamespaceMixin, InstanceFormMixin, DetailView):
    model = Recording
    template_name_suffix = "_form"

    def get_context_data(self, **kwargs):
        context = super(RecordingUpdate, self).get_context_data(**kwargs)
        context['recordingtimestamp_formset'] = RecordingTimestampFormSet(
            self.request.POST or None, instance=self.object)
        return context

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)  # Sigh
        recordingtimestamp_formset = context['recordingtimestamp_formset']

        recording = self.object

        for form in recordingtimestamp_formset:
            # in case we are processing any extra fields, which don't have
            # this set already
            form.fields['timestamp'].recording_start = recording.start_datetime

        if recordingtimestamp_formset.is_valid():
            recordingtimestamp_formset.save()

            recording.create_or_update_speeches(self.request.instance)

            # then delete the associated speeches, as Django can't
            # infer a cascade here
            for form in recordingtimestamp_formset.deleted_forms:
                try:
                    form.instance.speech.delete()
                except:
                    logger.info("Timestamp isn't linked to speech")

            return HttpResponseRedirect(recording.get_absolute_url())
        return self.render_to_response(context)


class RecordingAPICreate(InstanceFormMixin, JSONResponseMixin, CreateView):
    # View for RecordingAPIForm, to create a recording
    model = Recording
    form_class = RecordingAPIForm

    # Limit this view to POST requests, we don't want to show an HTML form for it
    http_method_names = ['post']

    success_url = 'DUMMY'

    def get_form_kwargs(self):
        kwargs = super(RecordingAPICreate, self).get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs

    def form_valid(self, form):
        logger.info("Processing recording")
        logger.info("  " + repr(form.cleaned_data))

        super(RecordingAPICreate, self).form_valid(form)

        recording = self.object

        recording_timestamps = form.cleaned_data.get('timestamps', [])
        for recording_timestamp in recording_timestamps:
            recording_timestamp.recording = recording
            recording_timestamp.save()

        if len(recording_timestamps):
            recording.start_datetime = recording_timestamps[0].timestamp
            recording.save()

        # Create speeches from the recording
        speeches = recording.create_or_update_speeches(self.request.instance)

        # Transcribe each speech
        for speech in speeches:
            speech.start_transcribing()

        # Return a 201 response
        serialisable_fields = ('audio', 'timestamps')
        serialised = serializers.serialize(
            "json", [recording], fields=serialisable_fields)
        serialised = serialised[1:-1]

        return self.render_to_response(
            serialised,
            status=201,
            location=reverse("speeches:recording-view", args=[recording.id]),
            )

    def form_invalid(self, form):
        return self.render_to_response(
            {'errors': json.dumps(form.errors)},
            status=400,
            )


class Select2AutoResponseView(AutoResponseView):
    def check_all_permissions(self, request, *args, **kwargs):
        super(Select2AutoResponseView, self).check_all_permissions(request, *args, **kwargs)

        try:
            id = int(request.GET['object_id'])
        except:
            id = None

        model = request._AutoResponseView__django_select2_local.model
        qs = model.objects.for_instance(request.instance)
        if id:
            qs = qs.exclude(id=id)
        request._AutoResponseView__django_select2_local.queryset = qs


class PopoloImportView(NamespaceMixin, InstanceFormMixin, FormView):
    template_name = 'speeches/popolo_import_form.html'
    form_class = PopoloImportForm

    def get_success_url(self):
        return self.reverse('speeches:speaker-list')

    def get_form_kwargs(self):
        kwargs = super(PopoloImportView, self).get_form_kwargs()
        kwargs['instance'] = self.request.instance
        return kwargs

    def form_valid(self, form):
        results = form.cleaned_data['importer'].import_persons()

        created = results['created']
        refreshed = results['refreshed']

        if created or refreshed:
            messages.add_message(
                self.request,
                messages.SUCCESS if created else messages.INFO,
                ' '.join((
                    ungettext(
                        "%(created)d speaker created.",
                        "%(created)d speakers created.",
                        created,
                        ) % {'created': created},
                    ungettext(
                        "%(refreshed)d speaker refreshed.",
                        "%(refreshed)d speakers refreshed.",
                        refreshed,
                        ) % {'refreshed': refreshed},
                    ))
                )

        else:
            messages.add_message(
                self.request,
                messages.WARNING,
                _('No speakers found.'),
                )

        return super(PopoloImportView, self).form_valid(form)


class AkomaNtosoImportView(NamespaceMixin, InstanceFormMixin, FormView):
    template_name = 'speeches/akoma_ntoso_import_form.html'
    form_class = AkomaNtosoImportForm

    def get_success_url(self):
        return self.reverse('speeches:home')

    def get_form_kwargs(self):
        kwargs = super(AkomaNtosoImportView, self).get_form_kwargs()
        kwargs['instance'] = self.request.instance
        return kwargs

    def form_valid(self, form):
        try:
            importer = ImportAkomaNtoso(
                instance=self.request.instance,
                commit=True,
                clobber=form.cleaned_data.get('existing_sections'),
                )

            stats = importer.import_document(form.cleaned_data['location'])
        except:
            form._errors[NON_FIELD_ERRORS] = form.error_class(
                [_('Sorry - something went wrong with the import')])
            return self.form_invalid(form)

        speakers = stats.get(Speaker, 0)
        sections = stats.get(Section, 0)
        speeches = stats.get(Speech, 0)
        success = speakers or sections or speeches

        if success:
            messages.add_message(
                self.request,
                messages.SUCCESS,
                _('Created: ') + ', '.join(
                    (
                        ungettext(
                            "%(speakers)d speaker",
                            "%(speakers)d speakers",
                            speakers,
                            ) % {'speakers': speakers},
                        ungettext(
                            "%(sections)d section",
                            "%(sections)d sections",
                            sections,
                            ) % {'sections': sections},
                        ungettext(
                            "%(speeches)d speech",
                            "%(speeches)d speeches",
                            speeches,
                            ) % {'speeches': speeches},
                        )
                    )
                )
        else:
            messages.add_message(
                self.request,
                messages.INFO,
                _('Nothing new to import.'),
                )

        return super(AkomaNtosoImportView, self).form_valid(form)



# API Resources
# Uses Tastypie to do most of the hard work
# http://django-tastypie.readthedocs.org/en/v0.9.11/tutorial.html
from tastypie import fields
from tastypie.resources import NamespacedModelResource, ALL_WITH_RELATIONS, ALL
from tastypie.authentication import Authentication
from tastypie.authorization import ReadOnlyAuthorization

from haystack.query import SearchQuerySet

from speeches.models import Speech, Speaker, Section


class SectionResource(NamespacedModelResource):
    parent = fields.ForeignKey('self', 'parent', null=True)
    children = fields.ToManyField('self', 'children')

    def get_object_list(self, request):
        return super(SectionResource, self).get_object_list(request).filter(instance=request.instance)

    class Meta:
        queryset = Section.objects.all()
        resource_name = 'section'
        excludes = []
        allowed_methods = ['get']
        authentication = Authentication()
        authorization = ReadOnlyAuthorization()
        filtering = {
            'num': ALL,
            'heading': ALL,
            'subheading': ALL,
            'parent': ALL,
        }


class SpeakerResource(NamespacedModelResource):
    def get_object_list(self, request):
        return super(SpeakerResource, self).get_object_list(request).filter(instance=request.instance)

    class Meta:
        queryset = Speaker.objects.all()
        resource_name = 'speaker'
        excludes = []
        allowed_methods = ['get']
        authentication = Authentication()
        authorization = ReadOnlyAuthorization()
        filtering = {
            'name': ALL,
        }


class SpeechResource(NamespacedModelResource):
    speaker = fields.ForeignKey(SpeakerResource, 'speaker', null=True, full=True)

    def apply_filters(self, request, applicable_filters):
        objects = super(SpeechResource, self).apply_filters(request, applicable_filters)
        if 'q' in request.GET:
            objects = objects.auto_query(request.GET['q'])
            objects = [result.object for result in objects]
        return objects

    def get_object_list(self, request):
        if 'q' in request.GET:
            return SearchQuerySet().models(Speech).narrow('instance:%s' % request.instance.label).load_all()
        return super(SpeechResource, self).get_object_list(request).filter(instance=request.instance)

    class Meta:
        queryset = Speech.objects.filter(public=True)
        resource_name = 'speech'
        excludes = ['celery_task_id', 'public']
        allowed_methods = ['get', 'post']
        authentication = Authentication()
        authorization = ReadOnlyAuthorization()
        filtering = {
            'speaker': ALL_WITH_RELATIONS,
            'start_date': ALL,
        }



import tempfile
import shutil

from datetime import datetime, date, timedelta

from django.test import TestCase
from django.test.utils import override_settings

from selenium.webdriver.support.ui import WebDriverWait

from speeches.models import Section, Speech
from instances.models import Instance

from instances.tests import InstanceTestCase, InstanceLiveServerTestCase


class InstanceTestCase(InstanceTestCase):
    default_instance_options = dict(label='default')


temp_media_root = tempfile.mkdtemp(prefix='sayit_test')


@override_settings(MEDIA_ROOT=temp_media_root)
class OverrideMediaRootMixin(TestCase):
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(temp_media_root)
        return super(OverrideMediaRootMixin, cls).tearDownClass()


class InstanceLiveServerTestCase(InstanceLiveServerTestCase):
    default_instance_options = dict(label='default')

    def setUp(self):
        super(InstanceLiveServerTestCase, self).setUp()
        WebDriverWait(self.selenium, 10).until(
            lambda x: x.find_element_by_link_text('Add your first statement')
        )


class ParentInstanceMismatchError(Exception):
    pass


def create_sections(subsections, parent=None, instance=None):
    """
    Create a hierachy of sections and speeches - good for setting up test data.

    Note - used in external packages (pombola.south_africa), alter with care.

    create_sections([
        {
            'heading': "Top level section",
            'subsections': [
                {   'heading': "Nested section",
                    'subsections': [
                        {   'heading': "Section with speeches",
                            'speeches': [ 4, date(2013, 3, 25), time(9, 0) ],
                        },
                        {   'heading': "Bill on Silly Walks",
                            'speeches': [ 2, date(2013, 3, 25), time(12, 0) ],
                        },
                    ]
                },
                {
                    'heading': "Another nested section (but completely empty)",
                    'subsections': []
                },
            ]
        },
    ])
    """

    # If given an instance and a parent check that they are compatible -
    # we don't want to end up with a hierarchy that spans instances and
    # this would probably mean that there is some confusion upstream that
    # we should highlight.
    if instance and parent and instance != parent.instance:
        raise ParentInstanceMismatchError(
            "The instance and parent.instance do not match")

    # If we have a parent then use the instance from that.
    if parent:
        instance = parent.instance

    # If we don't have an instance (which also means we had no parent)
    # then create one - this is very convenient in the test scripts.
    if not instance:
        instance, _ = Instance.objects.get_or_create(label='create-sections-instance')

    for subsection in subsections:
        s = Section.objects.create(
            instance=instance,
            heading=subsection['heading'],
            parent=parent,
            )
        if 'subsections' in subsection:
            create_sections(subsection['subsections'], parent=s)
        if 'speeches' in subsection:
            # If there's a 4th element in speeches, that's a boolean
            # indicating whether the speeches should be public:
            speeches_details = subsection['speeches']
            if len(speeches_details) >= 4:
                public = speeches_details[3]
            else:
                public = True
            num, d, t = speeches_details[:3]
            for i in range(0, num):
                Speech.objects.create(
                    instance=instance,
                    section=s,
                    text='rhubarb rhubarb',
                    start_date=d,
                    start_time=t,
                    public=public,
                    source_url="http://somewhere.or.other/{0}".format(i),
                )
                if t:
                    t = (datetime.combine(date.today(), t) + timedelta(minutes=10)).time()


class CreateSectionsTests(TestCase):
    def test_parent_instance_mismatch(self):
        foo_instance = Instance.objects.create(label="foo")
        bar_instance = Instance.objects.create(label="bar")
        foo_parent = Section.objects.create(
            instance=foo_instance,
            heading="Foo Section",
            )

        # Should run without exception
        create_sections([], parent=foo_parent, instance=foo_instance)
        create_sections([], instance=foo_instance)
        create_sections([], parent=foo_parent)
        create_sections([])

        # Should raise
        self.assertRaises(
            ParentInstanceMismatchError,
            create_sections,
            subsections=[],
            parent=foo_parent,
            instance=bar_instance,
        )



import calendar
import datetime
import hashlib
import logging
import os
from six.moves.urllib.parse import urlsplit
from six.moves.urllib.request import urlretrieve
from six.moves.urllib.error import HTTPError

from django.utils.translation import ugettext_lazy as _, ugettext
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.html import strip_tags
from django.template.defaultfilters import timesince
from django.conf import settings
from django.core.files import File
from django.core.exceptions import ValidationError
from django.contrib.contenttypes.fields import GenericRelation
from django.utils.encoding import python_2_unicode_compatible

from easy_thumbnails.fields import ThumbnailerImageField
from easy_thumbnails.alias import aliases

from instances.models import InstanceMixin, InstanceManager
from speeches.utils.audio import AudioHelper
from speeches.utils.text import url_to_unicode

from popolo.models import Person
from sluggable.fields import SluggableField
from sluggable.models import Slug as SlugModel

if not aliases.get('speaker-thumb'):
    aliases.set('speaker-thumb', {'size': (96, 96), 'crop': 'smart', 'upscale': True, 'face': True})

if not aliases.get('speaker-rectangle'):
    aliases.set('speaker-rectangle', {'size': (96, 0), 'upscale': True})

logger = logging.getLogger(__name__)

max_date = datetime.date(datetime.MAXYEAR, 12, 31)
max_time = datetime.time(23, 59)
max_datetime = datetime.datetime.combine(max_date, max_time)


class cache(object):
    '''Computes attribute value and caches it in the instance.
    Python Cookbook (Denis Otkidach) http://stackoverflow.com/users/168352/denis-otkidach
    This decorator allows you to create a property which can be computed once and
    accessed many times. Sort of like memoization.

    '''
    def __init__(self, method, name=None):
        # record the unbound-method and the name
        self.method = method
        self.name = name or method.__name__
        self.__doc__ = method.__doc__

    def __get__(self, inst, cls):
        # self: <__main__.cache object at 0xb781340c>
        # inst: <__main__.Foo object at 0xb781348c>
        # cls: <class '__main__.Foo'>
        if inst is None:
            # instance attribute accessed on class, return self
            # You get here if you write `Foo.bar`
            return self
        # compute, cache and return the instance's attribute value
        result = self.method(inst)
        # setattr redefines the instance's attribute so this doesn't get called again
        setattr(inst, self.name, result)
        return result


class AuditedModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        now = timezone.now()
        if not self.id:
            self.created = now
        self.modified = now
        super(AuditedModel, self).save(*args, **kwargs)


class Slug(SlugModel):
    pass


def upload_to(inst, fn):
    return inst.get_image_cache_file_path(fn)


class SpeakerManager(InstanceManager):
    def get_queryset(self):
        return super(SpeakerManager, self).get_queryset().extra(
            select={"sorted_name": "LOWER(COALESCE(NULLIF(sort_name, ''), name))"},
            order_by=("sorted_name",))


# Speaker - someone who gave a speech
@python_2_unicode_compatible
class Speaker(InstanceMixin, Person):
    slug = SluggableField(
        _('slug'), unique_with='instance', populate_from='name', always_update=True)
    slugs = GenericRelation(Slug)

    # There's an unfortunate collision of things called 'instance' here.
    # inst is the model instance, in this case a Speaker. inst.instance
    # is the django subdomain instance.
    image_cache = ThumbnailerImageField(
        _('image_cache'),
        upload_to=upload_to,
        null=True,
        blank=True,
        help_text=_('If image is set, a local copy will be stored here.'),
        )

    objects = SpeakerManager()

    class Meta:
        verbose_name = _('speaker')
        verbose_name_plural = _('speakers')
        unique_together = ('instance', 'slug')

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if self.image:
            try:
                tmp_filename, headers = urlretrieve(self.image)

                max_filename_length = self.image_cache.field.max_length  # Usually 100
                template_needs = len(self.image_cache_file_path_template % '')

                # We'll use the actual instance label size rather than the maximum that it could be
                # instance_label_length = self.instance._meta.get_field_by_name('label')[0].max_length
                instance_label_length = len(self.instance.label)
                truncate_to = max_filename_length - template_needs - instance_label_length - 8

                image_filename = os.path.basename(urlsplit(self.image).path)
                image_filename = url_to_unicode(image_filename)
                filename_root, extension = os.path.splitext(image_filename)

                truncated_filename = filename_root[:(truncate_to - len(extension))] + extension

                self.image_cache.save(
                    truncated_filename,
                    File(open(tmp_filename, 'rb')),

                    # Calling save on a FieldFile causes a save on the model instance
                    # we don't need that as we're about to do it below (without this
                    # we get an infinite recursion).
                    save=False,
                    )
            except HTTPError:
                pass

        return super(Speaker, self).save(*args, **kwargs)

    image_cache_file_path_template = 'speakers/%s/'

    def get_image_cache_file_path(self, filename):
        path = self.image_cache_file_path_template % (self.instance.label)
        return os.path.join(path, filename)

    @property
    def colour(self):
        id = ('%s' % self.person_ptr_id).encode()
        return hashlib.sha1(id).hexdigest()[:6]

    @models.permalink
    def get_absolute_url(self):
        return ('speeches:speaker-view', (), {'slug': self.slug})

    @models.permalink
    def get_edit_url(self):
        return ('speeches:speaker-edit', (), {'pk': self.person_ptr_id})


@python_2_unicode_compatible
class Tag(InstanceMixin, AuditedModel):
    name = models.CharField(unique=True, max_length=100)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _('tag')
        verbose_name_plural = _('tags')


class SectionManager(InstanceManager):

    def get_or_create_with_parents(self, instance, headings):
        """Get or create the entire hierarchy of sections given, returning the bottom one."""

        # If there are no headings to create we just return None
        if not len(headings):
            return None

        # Get the heading to create, the rest and get the parent
        heading = headings[-1]
        rest = headings[:-1]
        parent = self.get_or_create_with_parents(instance, rest)

        section, created = self.get_or_create(instance=instance, heading=heading, parent=parent)

        return section


@python_2_unicode_compatible
class Section(AuditedModel, InstanceMixin):
    # Custom manager
    objects = SectionManager()

    num = models.TextField(
        _('number'), blank=True, help_text=_('The number of the section'))
    heading = models.TextField(
        _('heading'), blank=True, help_text=_('The heading of the section'))
    subheading = models.TextField(
        _('subheading'), blank=True,
        help_text=_('The subheading of the section'))
    description = models.TextField(
        _('description'), blank=True, help_text=_('Longer description, HTML'))
    start_date = models.DateField(
        _('start date'), blank=True, null=True,
        help_text=_('What date did the section start?'))
    start_time = models.TimeField(
        _('start time'), blank=True, null=True,
        help_text=_('What time did the section start?'))
    number = models.TextField(_('document number'), blank=True)
    legislature = models.TextField(_('legislature'), blank=True)
    session = models.TextField(
        _('session'), blank=True, help_text=_('Legislative session'))
    parent = models.ForeignKey(
        'self', verbose_name=_('parent'), null=True, blank=True,
        related_name='children')
    slug = SluggableField(
        _('slug'), unique_with=('parent', 'instance'), populate_from='title', always_update=True)
    source_url = models.TextField(_('source URL'), blank=True)

    slugs = GenericRelation(Slug)

    class Meta:
        verbose_name = _('section')
        verbose_name_plural = _('sections')
        ordering = ('id',)
        unique_together = ('parent', 'slug', 'instance')

    def __str__(self):
        return self.heading or ugettext('Section')

    @property
    def title(self):
        txt = filter(None, (self.num, self.heading, self.subheading))
        # If there's nothing to set the title from, we should still have something
        # non-empty to make a slug out of.
        return ', '.join(txt) or 'Untitled'

    def clean(self):
        super(Section, self).clean()
        if not self.num and not self.heading and not self.subheading:
            raise ValidationError(_('You must specify at least one of num/heading/subheading'))

    def speech_datetimes(self):
        return (datetime.datetime.combine(s.start_date, s.start_time or datetime.time(0, 0))
                for s in self.speech_set.all())

    def is_leaf_node(self):
        return not self.children.exists()

    @cache
    def get_children(self):
        tree = self.get_descendants
        try:
            lvl = tree[0].level
            return [s for s in tree if s.level == lvl]
        except:
            return []

    @cache
    def get_ancestors(self):
        # Cached for now, so these aren't supplied as arguments
        ascending = False
        include_self = True
        """Return the ancestors of the current Section, in either direction,
        optionally including itself."""

        if not self.parent_id:
            # Shortcut when we know there's no parent
            s = [self]
        else:
            dir = ascending and 'ASC' or 'DESC'
            s = Section.objects.raw(
                """WITH RECURSIVE cte AS (
                    SELECT speeches_section.*, 1 AS l FROM speeches_section WHERE id=%s
                    UNION ALL
                    SELECT s.*, l+1 FROM cte JOIN speeches_section s ON cte.parent_id = s.id
                )
                SELECT * FROM cte ORDER BY l """ + dir,
                [self.id]
            )
        if not include_self:
            s = ascending and s[1:] or s[:-1]
        return list(s)  # So it's evaluated and will be cached

    def _get_descendants(self, include_self=False, include_count='', include_min='', max_depth=''):
        """Return the descendants of the current Section, in depth-first order.
        Optionally, include speech counts, minimum speech times, and only
        descend a certain depth."""
        select = ['*']
        if include_count:
            select.append("(SELECT COUNT(*) FROM speeches_speech WHERE section_id = cte.id) AS speech_count")
        if include_min:
            select.append("""(SELECT MIN( start_date + COALESCE(start_time, time '00:00') )
                FROM speeches_speech WHERE section_id = cte.id) AS speech_min""")
        if max_depth:
            max_depth = "WHERE array_upper(path, 1) < %d" % (max_depth + 1)
        s = Section.objects.raw(
            """WITH RECURSIVE cte AS (
                SELECT speeches_section.*, 0 AS level, ARRAY[id] AS path FROM speeches_section WHERE id=%s
                UNION ALL
                SELECT s.*, level+1, cte.path||s.id FROM cte JOIN speeches_section s ON cte.id = s.parent_id
            """ + max_depth + """
            )
            SELECT """ + ','.join(select) + """ FROM cte ORDER BY path""",
            [self.id]
        )
        if not include_self:
            s = s[1:]
        return s

    @cache
    def get_descendants_tree(self):
        """Given a flat list of Sections ordered by speech_min, create a
        hierarchical structure containing the same information."""
        d = self._get_descendants_by_speech(include_count=True)
        prev = self
        parents = []
        for node in d:
            if node.level > getattr(prev, 'level', 0):
                parents.append(prev)
                prev._childs = []
            elif node.level < prev.level:
                parents = parents[:node.level - prev.level]
            elif node.path[:-1] != prev.path[:-1]:  # Swapping parentage in some way
                sw = next((i for i in range(len(node.path)) if node.path[i] != prev.path[i]))
                parents = parents[:sw - node.level]
                for i in range(sw, node.level):
                    section = Section.objects.get(id=node.path[i])
                    section_copy = Section(
                        num=section.num,
                        heading=section.heading,
                        subheading=section.subheading,
                        parent=section.parent,
                    )
                    section_copy._childs = []
                    parents[-1]._childs.append(section_copy)
                    parents.append(section_copy)
            prev = node
            parents[-1]._childs.append(node)
        return d

    def get_descendants_tree_with_speeches(self, request, all_speeches=False):
        # Get the descendants tree of sections
        tree = self.get_descendants_tree

        # Fetch all speeches in this section, or all descendant sections
        section_list = [self]
        if all_speeches:
            section_list.extend(tree)

        speech_list = Speech.objects.filter(
            section__in=section_list).visible(request).select_related('speaker').prefetch_related('tags')
        self._speeches_by_section = {}
        for s in speech_list:
            self._speeches_by_section.setdefault(s.section_id, []).append(s)

        self._interleave_speeches(self)

        speech_type_choices = dict(Speech._meta.get_field('type').choices)

        # Finally, work back into a flat format for template display
        tree_final = []

        def rec(s, l):
            for i, c in enumerate(s._childs):
                attrs = {}
                if i == 0:
                    attrs['new_level'] = True
                if isinstance(c, Speech):
                    attrs['speech'] = True

                    # Anything with an odd type gets it replaced with 'other'
                    if c.type not in speech_type_choices:
                        c.type = 'other'

                tree_final.append((c, attrs))

                if isinstance(c, Section):
                    rec(c, l + 1)

                if i == len(s._childs) - 1:
                    tree_final[-1][1].setdefault('closed_levels', []).append(l)

        rec(self, 1)

        # Return an iterator because otherwise passing this down in context to
        # subtemplates is really slow. I mean, upwards of 30 seconds slow.
        # As a class so it can be reused after use (e.g. by the tests).
        class _iterable(object):
            def __iter__(self):
                return iter(tree_final)
        return _iterable()

    def _interleave_speeches(self, section):
        if not hasattr(section, '_childs'):
            section._childs = []

        # Recurse through the tree
        for child in section._childs:
            self._interleave_speeches(child)

        # Create a sorting key for each entry of the tree
        tree_with_key = [
            (
                (
                    getattr(d, 'speech_min', None) or max_datetime,
                    float('inf'),  # Always compares greater than speech ID
                    i
                ),
                d
            ) for i, d in enumerate(section._childs)
        ]

        # Create a sorting key for each speech
        speech_list_with_key = [
            (
                (
                    datetime.datetime.combine(
                        s.start_date or max_date, s.start_time or max_time
                    ),
                    s.id
                ),
                s
            ) for s in self._speeches_by_section.get(section.id, [])
        ]

        # Sort by our sorting keys to interleave the two
        tree_sorted = [s[1] for s in sorted(tree_with_key + speech_list_with_key)]
        section._childs = tree_sorted

    @cache
    def get_descendants(self):
        return self._get_descendants_by_speech()

    def _get_descendants_by_speech(self, **kwargs):
        dqs = self._get_descendants(include_min=True, **kwargs)

        max_datetime = datetime.datetime(datetime.MAXYEAR, 12, 31)

        lookup = dict((x.id, x) for x in dqs)
        lookup[self.id] = self
        self.speech_min = max_datetime

        for d in dqs:
            if not d.speech_min:
                continue
            for parent_id in reversed(d.path):
                parent = lookup[parent_id]
                if not parent.speech_min or d.speech_min < parent.speech_min:
                    parent.speech_min = d.speech_min

        # Set the speech_min to all the earliests
        for d in dqs:
            if d.speech_min:
                continue
            for parent_id in reversed(d.path):
                parent = lookup[parent_id]
                if parent.speech_min:
                    d.speech_min = parent.speech_min
                    break

        return sorted(dqs, key=lambda s: getattr(s, 'speech_min', max_datetime))

    @models.permalink
    def get_absolute_url(self):
        return ('speeches:section-view', (), {'full_slug': self.get_path})

    @models.permalink
    def get_edit_url(self):
        return ('speeches:section-edit', (), {'pk': self.id})

    @models.permalink
    def get_delete_url(self):
        return ('speeches:section-delete', (), {'pk': self.id})

    @cache
    def get_path(self):
        return '/'.join([s.slug for s in self.get_ancestors])

    def _get_next_previous_node(self, direction):
        if not self.parent:
            return None
        root = self.get_ancestors[0]
        tree = root.get_descendants
        idx = tree.index(self)
        lvl = tree[idx].level
        same_level = [s for s in tree if s.level == lvl]
        idx = same_level.index(self)
        if direction == -1 and idx == 0:
            return None
        try:
            return same_level[idx + direction]
        except:
            return None

    def get_next_node(self):
        """Fetch the next node in the tree, at the same level as this one."""
        return self._get_next_previous_node(1)

    def get_previous_node(self):
        """Fetch the previous node in the tree, at the same level as this one."""
        return self._get_next_previous_node(-1)

    def descendant_speeches(self):
        """Return a queryset of all speeches that belong to this section, or any of its descendants."""
        section_descendants = self._get_descendants(include_self=True)
        section_descendants_ids = [x.id for x in section_descendants]
        return Speech.objects.filter(section__in=section_descendants_ids)


class AudioMP3Mixin(object):
    def save(self, *args, **kwargs):
        """Overriden save method to automatically convert the audio to an mp3"""
        duration = kwargs.pop('duration', None)
        needs_conversion = self.audio and not self.audio.name.lower().endswith('.mp3')

        # If we have an audio file and it's not an mp3, make it one
        if duration or needs_conversion:
            if not os.path.exists(self.audio.path):
                # If it doesn't already exist, save the old audio first so that we can re-encode it
                # This is needed if it's newly uploaded
                self.audio.save(self.audio.name, File(self.audio), save=False)

            audio_helper = AudioHelper()

            if needs_conversion:
                mp3_filename = audio_helper.make_mp3(self.audio.path)
                mp3_file = open(mp3_filename, 'rb')
                self.audio.save(mp3_file.name, File(mp3_file), save=False)

            if duration:
                self.audio_duration = audio_helper.get_audio_duration(self.audio.path)

        # Call the original model save to do everything
        super(AudioMP3Mixin, self).save(*args, **kwargs)


# Speech manager
class SpeechQuerySet(models.QuerySet):
    def visible(query, request=None):
        if not request.is_user_instance:
            query = query.filter(public=True)
        return query


SpeechManager = InstanceManager.from_queryset(SpeechQuerySet)


# Speech that a speaker gave
@python_2_unicode_compatible
class Speech(InstanceMixin, AudioMP3Mixin, AuditedModel):
    # Custom manager
    objects = SpeechManager()

    # The speech. Need to check have at least one of the following two
    # (preferably both).
    audio = models.FileField(
        _('audio'), upload_to='speeches/%Y-%m-%d/', max_length=255, blank=True)
    text = models.TextField(
        _('text'), blank=True, db_index=False,
        help_text=_('The text of the speech'))

    # The section that this speech is part of
    section = models.ForeignKey(
        Section, verbose_name=_('Section'), blank=True, null=True,
        on_delete=models.SET_NULL,
        help_text=('The section that this speech is contained in'),)
    num = models.TextField(
        _('number'), blank=True,
        help_text=_('The number of the speech, if relevant'))
    heading = models.TextField(
        _('heading'), blank=True,
        help_text=_('The heading of the speech, if relevant'))
    subheading = models.TextField(
        _('subheading'), blank=True,
        help_text=_('The subheading of the speech, if relevant'))
    # The below two fields could be on the section if we made it a required
    # field of a speech
    event = models.TextField(
        _('event'), db_index=True, blank=True,
        help_text=_('Was the speech at a particular event?'))
    location = models.TextField(
        _('location'), db_index=True, blank=True,
        help_text=_('Where the speech took place'))

    # Metadata on the speech
    speaker = models.ForeignKey(
        Speaker, verbose_name=_('speaker'), blank=True, null=True,
        help_text=_('Who gave this speech?'), on_delete=models.SET_NULL)
    # may be null, in which case we simply use current strategy to get name
    speaker_display = models.CharField(max_length=256, null=True, blank=True)

    type = models.CharField(
        _('type'), max_length=32,
        # Expected values from Akoma Ntoso
        choices=(('speech', _('Speech')),
                 ('question', _('Question')),
                 ('answer', _('Answer')),
                 ('scene', _('Scene')),
                 ('narrative', _('Narrative')),
                 ('summary', _('Summary')),
                 ('other', _('Other')),
                 ),
        help_text=_('What sort of speech is this?'),
        )

    start_date = models.DateField(
        _('start date'), blank=True, null=True,
        help_text=_('What date did the speech start?'))
    start_time = models.TimeField(
        _('start time'), blank=True, null=True,
        help_text=_('What time did the speech start?'))
    end_date = models.DateField(
        _('end date'), blank=True, null=True,
        help_text=_('What date did the speech end?'))
    end_time = models.TimeField(
        _('end time'), blank=True, null=True,
        help_text=_('What time did the speech end?'))
    tags = models.ManyToManyField(
        Tag, verbose_name=_('tags'), blank=True)

    public = models.BooleanField(
        _('public'), default=True, help_text=_('Is this speech public?'))

    # What if source material has multiple speeches, same timestamp - need a
    # way of ordering them?
    # pos = models.IntegerField()

    source_url = models.TextField(_('source URL'), blank=True)
    # source_column? Any other source based things?

    # Task id for celery transcription tasks
    celery_task_id = models.CharField(
        _('celery task ID'), max_length=256, null=True, blank=True)

    class Meta:
        verbose_name = _('speech')
        verbose_name_plural = _('speeches')
        ordering = ('start_date', 'start_time', 'id')

    def __str__(self):
        out = 'Speech'
        for att in ('num', 'heading', 'subheading'):
            if getattr(self, att, None):
                out += u', %s' % getattr(self, att)
        if self.speaker:
            out += u' by %s' % self.speaker
        if self.start_date:
            out += ' at %s' % self.start_date
        if self.text:
            out += ' (with text)'
        if self.audio:
            out += ' (with audio)'
        return out

    @property
    def title(self):
        txt = filter(None, (self.num, self.heading, self.subheading))
        return ', '.join(txt)

    @property
    def is_public(self):
        return self.public

    @property
    def summary(self):
        summary_length = getattr(settings, 'SPEECH_SUMMARY_LENGTH', 30)
        if self.audio and not self.text:
            return "[ recorded audio ]"
        else:
            text = strip_tags(self.text)
            return text[:summary_length] + '...' if len(text) > summary_length else text

    @property
    def start_datetime(self):
        if self.start_date:
            return datetime.datetime.combine(self.start_date, self.start_time or datetime.time(0, 0))
        else:
            return None

    @property
    def end_datetime(self):
        if self.end_date:
            return datetime.datetime.combine(self.end_date, self.end_time or datetime.time(0, 0))
        else:
            return None

    @models.permalink
    def get_absolute_url(self):
        return ('speeches:speech-view', (), {'pk': self.id})

    @models.permalink
    def get_edit_url(self):
        return ('speeches:speech-edit', (), {'pk': self.id})

    @models.permalink
    def get_delete_url(self):
        return ('speeches:speech-delete', (), {'pk': self.id})

    def get_next_speech(self):
        """Return the next speech to this one in the same section, in a start
        date/time/ID ordering."""
        if self.start_date:
            # A later date, or no date present
            q1 = Q(start_date__gt=self.start_date) | Q(start_date__isnull=True)
            # Or, if on the same date...
            q2 = Q(start_date=self.start_date)
        else:
            # No dates later than our unknown
            q1 = Q()
            # But in that case...
            q2 = Q(start_date__isnull=True)
        if self.start_time:
            # If the time is later, or not present, or the same with a greater ID
            q3 = Q(start_time__gt=self.start_time) | Q(start_time__isnull=True) \
                | Q(start_time=self.start_time, id__gt=self.id)
        else:
            # If the time is not present and the ID is greater
            q3 = Q(start_time__isnull=True, id__gt=self.id)
        q = q1 | (q2 & q3)

        if self.section:
            s = list(Speech.objects.filter(q, section=self.section)[:1])
        else:
            s = list(Speech.objects.filter(q, instance=self.instance, section__isnull=True)[:1])
        if s:
            return s[0]
        if self.section:
            next_section = self.section.get_next_node()
            if next_section:
                s = next_section.speech_set.all()[:1]
        return s and s[0] or None

    def get_previous_speech(self):
        """Return the previous speech to this one in the same section,
        in a start date/time/ID ordering."""
        if self.start_date:
            # A date less than ours
            q1 = Q(start_date__lt=self.start_date)
            # Or if the same date...
            q2 = Q(start_date=self.start_date)
        else:
            # Any speech with a date is earlier
            q1 = Q(start_date__isnull=False)
            # And for those other speeches with no date...
            q2 = Q(start_date__isnull=True)
        if self.start_time:
            # If the time is earlier, or the same and the ID is smaller
            q3 = Q(start_time__lt=self.start_time) | Q(start_time=self.start_time, id__lt=self.id)
        else:
            # If there is a time, or there isn't but the ID is smaller
            q3 = Q(start_time__isnull=False) | Q(start_time__isnull=True, id__lt=self.id)
        q = q1 | (q2 & q3)

        if self.section:
            s = list(Speech.objects.filter(q, section=self.section).reverse()[:1])
        else:
            s = list(Speech.objects.filter(q, instance=self.instance, section__isnull=True).reverse()[:1])
        if s:
            return s[0]
        if self.section:
            prev_section = self.section.get_previous_node()
            if prev_section:
                s = prev_section.speech_set.all().reverse()[:1]
        return s and s[0] or None

    def start_transcribing(self):
        """In the past, kicked off a celery task to transcribe this speech"""
        return


# A timestamp of a particular speaker at a particular time.
# Used to record events like "This speaker started speaking at 00:33"
# in a specific recording, before it's chopped up into a speech
class RecordingTimestamp(InstanceMixin, AuditedModel):
    speaker = models.ForeignKey(Speaker, blank=True, null=True, on_delete=models.SET_NULL)
    timestamp = models.DateTimeField(db_index=True, blank=False)
    speech = models.ForeignKey(Speech, blank=True, null=True, on_delete=models.SET_NULL)
    recording = models.ForeignKey(
        'Recording', blank=False, null=False, related_name='timestamps',
        default=0)  # kludge default 0, should not be used

    class Meta:
        ordering = ('timestamp',)

    @property
    def utc(self):
        """Return our timestamp as a UTC long"""
        return calendar.timegm(self.timestamp.timetuple())


# A raw recording, might be divided up into multiple speeches
@python_2_unicode_compatible
class Recording(InstanceMixin, AudioMP3Mixin, AuditedModel):
    audio = models.FileField(upload_to='recordings/%Y-%m-%d/', max_length=255, blank=False)
    start_datetime = models.DateTimeField(
        blank=True, null=True, help_text='Datetime of first timestamp associated with recording')
    audio_duration = models.IntegerField(
        blank=True, null=False, default=0, help_text='Duration of recording, in seconds')

    def save(self, *args, **kwargs):
        kwargs['duration'] = True
        super(Recording, self).save(*args, **kwargs)

    def __str__(self):
        return u'Recording made %s ago' % timesince(self.created)

    @models.permalink
    def get_absolute_url(self):
        return ('speeches:recording-view', (), {'pk': self.id})

    def add_speeches_to_section(self, section):
        return Speech.objects.filter(recordingtimestamp__recording=self).update(section=section)

    def create_or_update_speeches(self, instance):
        created_speeches = []

        # Split the recording's audio files
        audio_helper = AudioHelper()
        audio_files = audio_helper.split_recording(self)
        sorted_timestamps = self.timestamps.order_by("timestamp")

        for index, audio_file in enumerate(audio_files):
            new = True
            speech = Speech(
                instance=instance,
                public=False,
                type='speech',
            )
            timestamp = None
            if sorted_timestamps and len(sorted_timestamps) > 0:
                # We assume that the files are returned in order of timestamp
                timestamp = sorted_timestamps[index]
                if timestamp.speech:
                    speech = timestamp.speech
                    new = False
                    try:
                        speech.audio.delete(save=False)
                    except:
                        pass
                        # shouldn't happen, but we're going to recreate anyway
                        # so not critical

                speech.speaker = timestamp.speaker
                speech.start_date = timestamp.timestamp.date()
                speech.start_time = timestamp.timestamp.time()
                # If there's another one we can work out the end too
                if index < len(sorted_timestamps) - 1:
                    next_timestamp = sorted_timestamps[index + 1]
                    speech.end_date = next_timestamp.timestamp.date()
                    speech.end_time = next_timestamp.timestamp.time()

            audio_file = open(audio_file, 'rb')
            speech.audio = File(audio_file)
            speech.save()

            if new:
                created_speeches.append(speech)
                if timestamp:
                    timestamp.speech = speech
                    timestamp.save()

        return created_speeches



# vim: set fileencoding=utf-8 :

import six

if six.PY2:
    from urllib import unquote as unquote_to_bytes
else:
    from urllib.parse import unquote_to_bytes


def url_to_unicode(s):
    # Python 2
    # >>> print(unquote('caf%c3%a9').decode('utf-8'))
    # 'café'
    # >>> print(unquote(u'caf%c3%a9').encode('raw_unicode_escape').decode('utf-8'))
    # 'café'
    #
    # Python 3
    # >>> print(unquote('caf%c3%a9'))
    # 'café'
    # >>> print(unquote_to_bytes(b'caf%c3%a9').decode('utf-8'))
    # 'café'
    #
    # So for consistent results, we need to encode() the incoming string
    # and call unquote (2) or unquote_to_bytes (3) before decode()
    s = s.encode('utf-8')
    s = unquote_to_bytes(s)
    return s.decode('utf-8')



import re
from mock import patch, Mock

import lxml.html

from django.utils.six import assertRegex
from django.utils.six.moves import urllib
from django.test.utils import override_settings

from speeches.tests import InstanceTestCase, OverrideMediaRootMixin
from speeches.models import Speaker, Speech, Section
from speeches import models

m = Mock()
m.return_value = ('speeches/fixtures/test_inputs/Ferdinand_Magellan.jpg', None)


@override_settings(MEDIA_URL='/uploads/')
class OpenGraphTests(OverrideMediaRootMixin, InstanceTestCase):
    @patch.object(models, 'urlretrieve', m)
    def setUp(self):
        super(OpenGraphTests, self).setUp()

        self.steve = Speaker.objects.create(
            name='Steve',
            instance=self.instance,
            image='http://example.com/image.jpg',
            )
        self.section = Section.objects.create(
            heading='Test section',
            instance=self.instance,
            )
        self.section_with_description = Section.objects.create(
            heading='Section with description',
            instance=self.instance,
            description='Section is described here',
            )
        self.section_with_html_description = Section.objects.create(
            heading='Bold Section',
            instance=self.instance,
            description='<b>Section is described <a href="/">here</a></b>',
            )
        self.steve_speech = Speech.objects.create(
            text="A Steve speech",
            instance=self.instance,
            speaker=self.steve,
            section=self.section,
            )
        self.speech_long_html_description = Speech.objects.create(
            text=(
                "<i>But I must explain to you how all this mistaken idea of "
                "denouncing pleasure and praising pain was born and I will "
                "give you a complete account of the system, and expound the "
                "actual teachings of the great explorer of the truth, the "
                "master-builder of human happiness.</i>"),
            instance=self.instance,
            speaker=self.steve,
            section=self.section,
            )

    def assert_opengraph_matches(self, response, graph):
        """Check that the response matches the graph.

        The graph should be passed in as a dictionary of opengraph
        property name to a content test. The test can be either
          * A string - this must be equal to the content of that key, or
          * A compliled regular expression - which must be found in the
            content for that OpenGraph key in a regex search.

        For example:
        {
            'title': 'OpenGraph title of the page',
            'url:': re.compile('http://testing.example.com:8000/images/*.jpg'),
        }

        Extra items in the graph in resp will be ignored - we're only
        checking that everything in graph appears correctly.
        """
        # Keys that should be URLs if they exist
        url_keys = set((
            'url', 'image', 'audio', 'video', 'image:url',
            'image:secure_url', 'video:secure_url', 'audio:secure_url'))

        parser = lxml.html.HTMLParser(encoding='utf-8')
        root = lxml.html.fromstring(response.content, parser=parser)

        for key, test in graph.items():
            content = root.xpath(".//meta[@property='og:%s']/@content" % key)
            self.assertEqual(len(content), 1)
            content = content[0]

            if hasattr(test, 'pattern'):
                assertRegex(self, content, test)
            else:
                self.assertEqual(content, test)

            if key in url_keys:
                # Check that the url is absolute.
                self.assertTrue(
                    urllib.parse.urlsplit(content).netloc,
                    'og:%s must be an absolute URL, not %s' % (key, content)
                    )

    def test_default_instance_homepage(self):
        self.assert_opengraph_matches(
            self.client.get('/'),
            {'title': 'SayIt',
             'url': 'http://testing.example.org:8000/',
             'site_name': 'SayIt',
             'description': 'Transcripts for the modern internet',
             'type': 'website',
             'image': 'http://testing.example.org:8000/static/speeches/img/apple-touch-icon-152x152.png',
             }
            )

    def test_speaker_detail_page(self):
        self.assert_opengraph_matches(
            self.client.get('/speaker/%s' % self.steve.slug),
            {'title': 'View Speaker: Steve :: SayIt',
             'url': 'http://testing.example.org:8000/speaker/%s' % self.steve.slug,
             'site_name': 'SayIt',
             'description': 'Speeches by Steve',
             'type': 'website',
             'image': re.compile('http://testing.example.org:8000/uploads/speakers/default/image.*.jpg'),
             }
            )

    def test_speech_detail_page(self):
        self.assert_opengraph_matches(
            self.client.get('/speech/%s' % self.steve_speech.id),
            {'title': u'\u201cA Steve speech\u201d :: SayIt',
             'url': 'http://testing.example.org:8000/speech/%s' % self.steve_speech.id,
             'site_name': 'SayIt',
             'description': 'A Steve speech',
             'type': 'website',
             'image': re.compile('http://testing.example.org:8000/uploads/speakers/default/image.*.jpg'),
             }
            )

        self.assert_opengraph_matches(
            self.client.get('/speech/%s' % self.speech_long_html_description.id),
            {'title': u'\u201cBut I must explain to you how ...\u201d :: SayIt',
             'url': 'http://testing.example.org:8000/speech/%s' % self.speech_long_html_description.id,
             'site_name': 'SayIt',
             'description': (
                 'But I must explain to you how all this mistaken idea of '
                 'denouncing pleasure and praising pain was born and I will '
                 'give you a complete account of the system, ...'),
             'type': 'website',
             'image': re.compile('http://testing.example.org:8000/uploads/speakers/default/image.*.jpg'),
             }
            )

    def test_section_detail_page(self):
        self.assert_opengraph_matches(
            self.client.get('/section/%s' % self.section.id),
            {'title': 'View Section: Test section :: SayIt',
             'url': 'http://testing.example.org:8000/section/%s' % self.section.id,
             'site_name': 'SayIt',
             'description': 'Transcripts for the modern internet',
             'type': 'website',
             'image': 'http://testing.example.org:8000/static/speeches/img/apple-touch-icon-152x152.png',
             }
            )

    def test_section_with_description_detail_page(self):
        self.assert_opengraph_matches(
            self.client.get('/section/%s' % self.section_with_description.id),
            {'title': 'View Section: Section with description :: SayIt',
             'url': 'http://testing.example.org:8000/section/%s' % self.section_with_description.id,
             'site_name': 'SayIt',
             'description': 'Section is described here',
             'type': 'website',
             'image': 'http://testing.example.org:8000/static/speeches/img/apple-touch-icon-152x152.png',
             }
            )

    def test_section_with_html_description_detail_page(self):
        self.assert_opengraph_matches(
            self.client.get('/section/%s' % self.section_with_html_description.id),
            {'title': 'View Section: Bold Section :: SayIt',
             'url': 'http://testing.example.org:8000/section/%s' % self.section_with_html_description.id,
             'site_name': 'SayIt',
             'description': 'Section is described here',
             'type': 'website',
             'image': 'http://testing.example.org:8000/static/speeches/img/apple-touch-icon-152x152.png',
             }
            )



from django.contrib import admin
from django.db import models

from speeches.models import Speaker, Speech, Section, Tag
from speeches.widgets import AudioFileInput


class SpeechAdmin(admin.ModelAdmin):
    date_hierarchy = 'start_date'
    list_filter = ('tags',)
    formfield_overrides = {
        models.FileField: {'widget': AudioFileInput},
    }


class SectionAdmin(admin.ModelAdmin):
    search_fields = ('num', 'heading', 'subheading')
    prepopulated_fields = {'slug': ('num', 'heading', 'subheading')}


admin.site.register(Section, SectionAdmin)
admin.site.register(Speaker)
admin.site.register(Speech, SpeechAdmin)
admin.site.register(Tag)

# class FooBarAdmin(admin.ModelAdmin):
#     prepopulated_fields = {"slug": ["name"]}
#     list_display  = [ 'slug', 'name', ]
#     search_fields = ['name']
#
# admin.site.register( FooBar, FooBarAdmin )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('speeches', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='speaker',
            options={'verbose_name': 'speaker', 'verbose_name_plural': 'speakers'},
        ),
        migrations.AlterField(
            model_name='recording',
            name='instance',
            field=models.ForeignKey(verbose_name='instance', to='instances.Instance'),
        ),
        migrations.AlterField(
            model_name='recordingtimestamp',
            name='instance',
            field=models.ForeignKey(verbose_name='instance', to='instances.Instance'),
        ),
        migrations.AlterField(
            model_name='section',
            name='instance',
            field=models.ForeignKey(verbose_name='instance', to='instances.Instance'),
        ),
        migrations.AlterField(
            model_name='speaker',
            name='instance',
            field=models.ForeignKey(verbose_name='instance', to='instances.Instance'),
        ),
        migrations.AlterField(
            model_name='speech',
            name='instance',
            field=models.ForeignKey(verbose_name='instance', to='instances.Instance'),
        ),
        migrations.AlterField(
            model_name='speech',
            name='tags',
            field=models.ManyToManyField(to='speeches.Tag', verbose_name='tags', blank=True),
        ),
        migrations.AlterField(
            model_name='tag',
            name='instance',
            field=models.ForeignKey(verbose_name='instance', to='instances.Instance'),
        ),
    ]



from django import forms

from haystack.forms import SearchForm
from haystack.views import SearchView
from haystack.query import SearchQuerySet

from speeches.models import Speaker, Speech, Section


class HMSearchForm(SearchForm):
    def search(self):
        sqs = super(HMSearchForm, self).search()
        sqs = sqs.models(*self.model)
        sqs = sqs.highlight()
        return sqs


class SpeechForm(HMSearchForm):
    """
    A form with a hidden integer field that searches the speaker ID field
    """
    p = forms.IntegerField(required=False, widget=forms.HiddenInput())

    model = [Speech, Section]

    def search(self):
        sqs = super(SpeechForm, self).search()
        if self.is_valid() and self.cleaned_data.get('p'):
            sqs = sqs.filter(speaker=self.cleaned_data['p'])
            try:
                self.speaker = Speaker.objects.get(id=self.cleaned_data['p'])
            except:
                pass
        return sqs


class SpeakerForm(HMSearchForm):
    model = [Speaker]


class InstanceSearchView(SearchView):
    """
    A subclass that filters the search query set to speeches within the current
    request's instace.
    """
    def __init__(self, *args, **kwargs):
        kwargs['form_class'] = SpeechForm
        super(InstanceSearchView, self).__init__(*args, **kwargs)

    def build_form(self, *args, **kwargs):
        sqs = SearchQuerySet()
        sqs = sqs.narrow('instance:"%s"' % self.request.instance.label)
        self.searchqueryset = sqs
        return super(InstanceSearchView, self).build_form(*args, **kwargs)

    def extra_context(self):
        if not self.query:
            return {}

        self.form_class = SpeakerForm
        person_form = self.build_form()
        return {
            'speaker_results': person_form.search(),
        }



from django.views.generic.detail import SingleObjectMixin
from django.http import HttpResponseRedirect, Http404

from speeches.utils.base32 import MistypedIDException, base32_to_int


class UnmatchingSlugException(Exception):
    pass


class Base32SingleObjectMixin(SingleObjectMixin):
    """Assumes pk is in base32, and decodes that before passing on to the
    parent to be resolved."""

    def get(self, request, *args, **kwargs):
        try:
            return super(Base32SingleObjectMixin, self).get(request, *args, **kwargs)
        except UnmatchingSlugException as e:
            return HttpResponseRedirect(e.args[0].get_absolute_url())

    def get_context_data(self, **kwargs):
        """Copy of 1.6's SingleObjectMixin, then we skip
        SingleObjectMixin's own class as it's broken in 1.4."""
        context = {}
        if self.object:
            context['object'] = self.object
            context_object_name = self.get_context_object_name(self.object)
            if context_object_name:
                context[context_object_name] = self.object
        context.update(kwargs)
        return super(SingleObjectMixin, self).get_context_data(**context)

    def get_object(self, queryset=None):
        pk = self.kwargs.get(self.pk_url_kwarg, None)
        mistyped = False
        if pk is not None:
            try:
                pk = base32_to_int(pk)
            except MistypedIDException as e:
                mistyped = True
                pk = e.args[0]
            except:
                raise Http404('Could not match id %s' % pk)
            self.kwargs[self.pk_url_kwarg] = pk
        obj = super(Base32SingleObjectMixin, self).get_object(queryset)
        if (self.kwargs['slug'] != obj.slug) or mistyped:
            raise UnmatchingSlugException(obj)
        return obj



import os
import tempfile
import shutil

from django.test.utils import override_settings
from django.conf import settings

import speeches
from speeches.models import Speech, Speaker, Recording
from speeches.utils.audio import AudioHelper
from speeches.tests import InstanceTestCase

import logging
logging.disable(logging.WARNING)


@override_settings(MEDIA_ROOT=tempfile.mkdtemp())
class RecordingTimestampTests(InstanceTestCase):

    # TODO refactor with RecordingAPITests
    @classmethod
    def setUpClass(cls):
        cls._in_fixtures = os.path.join(os.path.abspath(speeches.__path__[0]), 'fixtures', 'test_inputs')
        super(RecordingTimestampTests, cls).setUpClass()

    def tearDown(self):
        # Clear the recordings folder if it exists
        recordings_folder = os.path.join(settings.MEDIA_ROOT, 'recordings')
        if(os.path.exists(recordings_folder)):
            shutil.rmtree(recordings_folder)

    def test_modify_timestamps(self, filename='lamb.mp3'):
        audio = open(os.path.join(self._in_fixtures, filename), 'rb')

        SPEECHES = 3
        # NOTE: this test is loosely cargo-culted from recording_api_tests,
        # however here timestamps are hardcoded for simplicity
        resp = self.client.post('/api/v0.1/recording/', {
            'audio': audio,
            'timestamps': '[{"timestamp":0},{"timestamp":2000},{"timestamp":4000}]'
        })

        def check_response(resp, expected_code=201, expected_type='application/json'):
            recording = Recording.objects.order_by('-id')[0]
            # Check response headers
            self.assertEqual(resp.status_code, expected_code)
            self.assertEqual(resp['Content-Type'], expected_type)
            self.assertIn('/recording/%d' % recording.id, resp['Location'])
            return recording

        recording = check_response(resp)
        timestamps = recording.timestamps.all()

        # Check the right number of timestamps and speeches were made
        self.assertEqual(recording.timestamps.count(), SPEECHES)
        self.assertEqual(Speech.objects.count(), SPEECHES)

        def check_audio_durations(recording, durations):
            audio_helper = AudioHelper()
            for (rt, d) in zip(recording.timestamps.all(), durations):
                self.assertEqual(
                    audio_helper.get_audio_duration(rt.speech.audio.path),
                    d)

        check_audio_durations(recording, [2, 2, 1])

        form_data = {
            'timestamps-TOTAL_FORMS': 4,
            'timestamps-INITIAL_FORMS': 3,
            'timestamps-MAX_NUM_FORMS': 1000,
            'timestamps-0-recording': recording.id,
            'timestamps-0-id': timestamps[0].id,
            'timestamps-0-speaker': "",
            'timestamps-0-timestamp': 0,
            'timestamps-1-recording': recording.id,
            'timestamps-1-id': timestamps[1].id,
            'timestamps-1-speaker': "",
            'timestamps-1-timestamp': 3,
            'timestamps-2-recording': recording.id,
            'timestamps-2-id': timestamps[2].id,
            'timestamps-2-speaker': "",
            'timestamps-2-timestamp': 4,
            'timestamps-3-recording': recording.id,
            'timestamps-3-id': "",
            'timestamps-3-speaker': "",
            'timestamps-3-timestamp': "",
        }
        timestamps_2_id = timestamps[2].id
        resp = self.client.post(
            '/recording/%d/edit' % recording.id, form_data)
        recording = check_response(resp, 302, 'text/html; charset=utf-8')
        check_audio_durations(recording, [3, 1, 1])

        form_data['timestamps-1-DELETE'] = 'on'
        resp = self.client.post(
            '/recording/%d/edit' % recording.id, form_data)
        recording = check_response(resp, 302, 'text/html; charset=utf-8')
        check_audio_durations(recording, [4, 1])

        form_data = {
            'timestamps-TOTAL_FORMS': 2,
            'timestamps-INITIAL_FORMS': 2,
            'timestamps-MAX_NUM_FORMS': 1000,
            'timestamps-0-DELETE': 'on',
            'timestamps-0-recording': recording.id,
            'timestamps-0-id': timestamps[0].id,
            'timestamps-0-speaker': "",
            'timestamps-0-timestamp': 0,
            'timestamps-1-recording': recording.id,
            'timestamps-1-id': timestamps_2_id,
            'timestamps-1-speaker': "",
            'timestamps-1-timestamp': 1,
        }
        resp = self.client.post(
            '/recording/%d/edit' % recording.id, form_data)
        recording = check_response(resp, 302, 'text/html; charset=utf-8')
        check_audio_durations(recording, [4])

        # test new timestamp AFTER
        form_data = {
            'timestamps-TOTAL_FORMS': 2,
            'timestamps-INITIAL_FORMS': 1,
            'timestamps-MAX_NUM_FORMS': 1000,
            'timestamps-0-recording': recording.id,
            'timestamps-0-id': timestamps_2_id,
            'timestamps-0-speaker': "",
            'timestamps-0-timestamp': 1,
            'timestamps-1-recording': recording.id,
            'timestamps-1-id': '',
            'timestamps-1-speaker': "",
            'timestamps-1-timestamp': 4,
        }
        resp = self.client.post(
            '/recording/%d/edit' % recording.id, form_data)
        recording = check_response(resp, 302, 'text/html; charset=utf-8')
        check_audio_durations(recording, [3, 1])

        # test new timestamp BETWEEN
        form_data = {
            'timestamps-TOTAL_FORMS': 3,
            'timestamps-INITIAL_FORMS': 2,
            'timestamps-MAX_NUM_FORMS': 1000,
            'timestamps-0-recording': recording.id,
            'timestamps-0-id': timestamps_2_id,
            'timestamps-0-speaker': "",
            'timestamps-0-timestamp': 1,
            'timestamps-1-recording': recording.id,
            'timestamps-1-id': recording.timestamps.all()[1].id,
            'timestamps-1-speaker': "",
            'timestamps-1-timestamp': 4,
            'timestamps-2-recording': recording.id,
            'timestamps-2-id': '',
            'timestamps-2-speaker': "",
            'timestamps-2-timestamp': 3,
        }
        resp = self.client.post(
            '/recording/%d/edit' % recording.id, form_data)
        recording = check_response(resp, 302, 'text/html; charset=utf-8')
        check_audio_durations(recording, [2, 1, 1])

        # test new timestamp BEFORE
        form_data = {
            'timestamps-TOTAL_FORMS': 4,
            'timestamps-INITIAL_FORMS': 3,
            'timestamps-MAX_NUM_FORMS': 1000,
            'timestamps-0-recording': recording.id,
            'timestamps-0-id': timestamps_2_id,
            'timestamps-0-speaker': "",
            'timestamps-0-timestamp': 1,
            'timestamps-1-recording': recording.id,
            'timestamps-1-id': recording.timestamps.all()[1].id,
            'timestamps-1-speaker': "",
            'timestamps-1-timestamp': 3,
            'timestamps-2-recording': recording.id,
            'timestamps-2-id': recording.timestamps.all()[2].id,
            'timestamps-2-speaker': "",
            'timestamps-2-timestamp': 4,
            'timestamps-3-recording': recording.id,
            'timestamps-3-id': '',
            'timestamps-3-speaker': "",
            'timestamps-3-timestamp': 0,
        }
        resp = self.client.post(
            '/recording/%d/edit' % recording.id, form_data)
        recording = check_response(resp, 302, 'text/html; charset=utf-8')
        check_audio_durations(recording, [1, 2, 1, 1])

    def test_can_get_form(self, filename='lamb.mp3'):
        audio = open(os.path.join(self._in_fixtures, filename), 'rb')

        # NOTE: this test is loosely cargo-culted from recording_api_tests,
        # however here timestamps are hardcoded for simplicity
        resp = self.client.post('/api/v0.1/recording/', {
            'audio': audio,
            'timestamps': '[{"timestamp":0},{"timestamp":2000},{"timestamp":4000}]'
        })

        recording = Recording.objects.order_by('-id')[0]
        self.assertEqual(resp.status_code, 201)

        resp = self.client.get('/recording/%d/edit' % recording.id)
        self.assertEqual(resp.status_code, 200)

    def test_change_speaker(self, filename='lamb.mp3'):
        audio = open(os.path.join(self._in_fixtures, filename), 'rb')

        # NOTE: this test is loosely cargo-culted from recording_api_tests,
        # however here timestamps are hardcoded for simplicity
        resp = self.client.post('/api/v0.1/recording/', {
            'audio': audio,
            'timestamps': '[{"timestamp":0},{"timestamp":2000},{"timestamp":4000}]'
        })

        recording = Recording.objects.order_by('-id')[0]
        self.assertEqual(resp.status_code, 201)

        resp = self.client.get('/recording/%d/edit' % recording.id)
        self.assertEqual(resp.status_code, 200)

        speaker_1 = Speaker.objects.create(name='Steve', instance=self.instance)
        speaker_2 = Speaker.objects.create(name='Yasmin', instance=self.instance)

        timestamps = recording.timestamps.all()

        form_data = {
            'timestamps-TOTAL_FORMS': 2,
            'timestamps-INITIAL_FORMS': 2,
            'timestamps-MAX_NUM_FORMS': 1000,
            'timestamps-0-recording': recording.id,
            'timestamps-0-id': timestamps[0].id,
            'timestamps-0-speaker': speaker_1.id,
            'timestamps-0-timestamp': 0,
            'timestamps-1-recording': recording.id,
            'timestamps-1-id': timestamps[1].id,
            'timestamps-1-speaker': speaker_2.id,
            'timestamps-1-timestamp': 1,
        }
        resp = self.client.post(
            '/recording/%d/edit' % recording.id, form_data)
        recording = Recording.objects.order_by('-id')[0]

        timestamps = recording.timestamps.all()
        self.assertEqual(timestamps[0].speaker, speaker_1)
        self.assertEqual(timestamps[1].speaker, speaker_2)



import json
import os
import tempfile
import shutil
from datetime import datetime
import pytz

from django.test.utils import override_settings
from django.conf import settings

import speeches
from speeches.models import Speech, Speaker, Recording
from speeches.tests import InstanceTestCase


@override_settings(MEDIA_ROOT=tempfile.mkdtemp())
class RecordingAPITests(InstanceTestCase):

    @classmethod
    def setUpClass(cls):
        cls._in_fixtures = os.path.join(os.path.abspath(speeches.__path__[0]), 'fixtures', 'test_inputs')
        super(RecordingAPITests, cls).setUpClass()

    def tearDown(self):
        # Clear the recordings folder if it exists
        recordings_folder = os.path.join(settings.MEDIA_ROOT, 'recordings')
        if(os.path.exists(recordings_folder)):
            shutil.rmtree(recordings_folder)

    def test_add_recording_fails_on_empty_form(self):
        # Test that the form won't submit if empty
        resp = self.client.post('/api/v0.1/recording/')
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp['Content-Type'], 'application/json')
        response_content = json.loads(resp.content.decode())
        self.assertEqual(response_content['errors'], '{"audio": ["This field is required."]}')

    def test_add_recording_with_audio(self):
        # Load the mp3 fixture
        audio = open(os.path.join(self._in_fixtures, 'lamb.mp3'), 'rb')

        resp = self.client.post('/api/v0.1/recording/', {
            'audio': audio
        })

        recording = Recording.objects.order_by('-id')[0]

        # Check response headers
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp['Content-Type'], 'application/json')
        self.assertIn('/recording/%d' % recording.id, resp['Location'])

        # Check response JSON
        response_content = json.loads(resp.content.decode())
        self.assertTrue(".mp3" in response_content['fields']['audio'])

        # Check in db
        self.assertIsNotNone(recording.audio)
        self.assertEqual(recording.audio_duration, 5)

    def test_add_recording_with_timestamp(self):
        # Add two speakers
        Speaker.objects.create(name='Steve', instance=self.instance)

        audio = open(os.path.join(self._in_fixtures, 'lamb.mp3'), 'rb')

        resp = self.client.post('/api/v0.1/recording/', {
            'audio': audio,
            'timestamps': '[{"speaker":"1","timestamp":946684800000}]'
        })

        recording = Recording.objects.order_by('-id')[0]

        # Check response headers
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp['Content-Type'], 'application/json')
        self.assertIn('/recording/%d' % recording.id, resp['Location'])

        # Check response JSON
        response_content = json.loads(resp.content.decode())
        self.assertTrue(".mp3" in response_content['fields']['audio'])

        # Check in db
        self.assertIsNotNone(recording.audio)
        self.assertEqual(recording.timestamps.count(), 1)
        expected_timestamp = datetime.utcfromtimestamp(946684800).replace(tzinfo=pytz.utc)
        self.assertEqual(recording.timestamps.all()[0].timestamp, expected_timestamp)
        self.assertEqual(recording.start_datetime, expected_timestamp)

    def test_add_ogg_with_multiple_timestamps(self):
        self.test_add_recording_with_multiple_timestamps('lamb.ogg')

    def test_add_recording_with_multiple_timestamps(self, filename='lamb.mp3'):
        SPEECHES = 3

        speakers = []
        expected_timestamps = []
        timestamps = '['
        for i in range(SPEECHES):
            s = Speaker.objects.create(name='Speaker ' + str(i), instance=self.instance)
            t = 946684800 + i * 3
            speakers.append(s)
            expected_timestamps.append(datetime.utcfromtimestamp(t).replace(tzinfo=pytz.utc))
            timestamps += '{"speaker":"' + str(s.id) + '","timestamp":' + str(t) + '000}'
            if i < SPEECHES - 1:
                timestamps += ','
        timestamps += ']'

        audio = open(os.path.join(self._in_fixtures, filename), 'rb')
        resp = self.client.post('/api/v0.1/recording/', {
            'audio': audio,
            'timestamps': timestamps
        })

        recording = Recording.objects.order_by('-id')[0]

        # Check response headers
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp['Content-Type'], 'application/json')
        self.assertIn('/recording/%d' % recording.id, resp['Location'])

        # Check response JSON
        response_content = json.loads(resp.content.decode())
        self.assertTrue(".mp3" in response_content['fields']['audio'])

        # Check in db
        # Check the recording
        self.assertIsNotNone(recording.audio)
        self.assertRegexpMatches(recording.audio.path, r'\.mp3$')

        # Check the right number of timestamps and speeches were made
        self.assertEqual(recording.timestamps.count(), SPEECHES)
        self.assertEqual(Speech.objects.count(), SPEECHES)

        # Check the timestamps and speeches are what we expect (ish)
        ordered_timestamps = recording.timestamps.order_by("timestamp")
        for i in range(SPEECHES):
            self.assertEqual(ordered_timestamps[i].timestamp, expected_timestamps[i])
            speech = Speech.objects.all()[i]
            self.assertEqual(speech.speaker, speakers[i])
            self.assertEqual(speech.start_date, expected_timestamps[i].date())
            self.assertEqual(speech.start_time, expected_timestamps[i].time())
            if i == SPEECHES - 1:
                self.assertEqual(speech.end_date, None)
                self.assertEqual(speech.end_time, None)
            else:
                self.assertEqual(speech.end_date, expected_timestamps[i + 1].date())
                self.assertEqual(speech.end_time, expected_timestamps[i + 1].time())
            self.assertIsNotNone(speech.audio.path)
            self.assertRegexpMatches(speech.audio.path, r'\.mp3$')
            self.assertEqual(speech, ordered_timestamps[i].speech)

    def test_add_recording_with_unknown_speaker_timestamp(self):
        audio = open(os.path.join(self._in_fixtures, 'lamb.mp3'), 'rb')

        expected_timestamp = datetime.utcfromtimestamp(946684800).replace(tzinfo=pytz.utc)

        resp = self.client.post('/api/v0.1/recording/', {
            'audio': audio,
            'timestamps': '[{"speaker":"", "timestamp":946684800000}]'
        })

        recording = Recording.objects.order_by('-id')[0]

        # Check response headers
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp['Content-Type'], 'application/json')
        self.assertIn('/recording/%d' % recording.id, resp['Location'])

        # Check response JSON
        response_content = json.loads(resp.content.decode())
        self.assertTrue(".mp3" in response_content['fields']['audio'])

        # Check in db
        self.assertIsNotNone(recording.audio)
        self.assertEqual(recording.timestamps.all().count(), 1)
        self.assertEqual(recording.timestamps.all()[0].timestamp, expected_timestamp)

    def test_add_recording_fails_with_unsupported_audio(self):
        # Load the .aiff fixture
        audio = open(os.path.join(self._in_fixtures, 'lamb.aiff'), 'rb')

        resp = self.client.post('/api/v0.1/recording/', {
            'audio': audio
        })

        # Assert that it fails and gives us an error
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp['Content-Type'], 'application/json')
        response_content = json.loads(resp.content.decode())
        response_errors = json.loads(response_content['errors'])
        self.assertEqual(response_errors['audio'], ["That file does not appear to be an audio file"])



# easy_thumbnails face cropping processor
# Much of the below taken from http://stackoverflow.com/a/13243712/669631

try:
    import cv
    faceCascade = cv.Load('/usr/share/opencv/haarcascades/haarcascade_frontalface_alt.xml')
except:
    faceCascade = False

# Select one of the haarcascade files:
#   haarcascade_frontalface_alt.xml  <-- Best one?
#   haarcascade_frontalface_alt2.xml
#   haarcascade_frontalface_alt_tree.xml
#   haarcascade_frontalface_default.xml
#   haarcascade_profileface.xml


def detectFaces(im):
    # This function takes a PIL image and finds the patterns defined in the
    # haarcascade function modified from: http://www.lucaamore.com/?p=638

    # Convert a PIL image to a greyscale cv image
    # from: http://pythonpath.wordpress.com/2012/05/08/pil-to-opencv-image/
    im = im.convert('L')
    cv_im = cv.CreateImageHeader(im.size, cv.IPL_DEPTH_8U, 1)
    cv.SetData(cv_im, im.tostring(), im.size[0])

    # variables
    min_size = (20, 20)
    haar_scale = 1.1
    min_neighbors = 3
    haar_flags = 0

    # Equalize the histogram
    cv.EqualizeHist(cv_im, cv_im)

    # Detect the faces
    faces = cv.HaarDetectObjects(
        cv_im, faceCascade, cv.CreateMemStorage(0),
        haar_scale, min_neighbors, haar_flags, min_size
        )

    return faces


def face_crop(im, size, face=False, **kwargs):
    if not face or not faceCascade:
        return im

    source_x, source_y = [int(v) for v in im.size]

    faces = detectFaces(im)
    if faces:
        cropBox = [0, 0, 0, 0]
        for face, n in faces:
            if face[2] > cropBox[2] or face[3] > cropBox[3]:
                cropBox = face

        xDelta = int(max(cropBox[2] * 0.25, 0))
        yDelta = int(max(cropBox[3] * 0.25, 0))

        # Convert cv box to PIL box [left, upper, right, lower]
        box = [
            max(cropBox[0] - xDelta, 0),
            max(cropBox[1] - yDelta, 0),
            min(cropBox[0] + cropBox[2] + xDelta, source_x - 1),
            min(cropBox[1] + cropBox[3] + yDelta, source_y - 1)
            ]
        im = im.crop(box)

    return im



import json
import os
import re
import logging
from datetime import datetime
import pytz

from django_select2.widgets import (
    Select2MultipleWidget, AutoHeavySelect2Widget,
    HeavySelect2Widget
    )
from django_select2.fields import AutoModelSelect2Field

from django.utils.translation import ugettext_lazy as _, ungettext
from django.utils.html import linebreaks

try:
    # Hack for Django 1.4 compatibility as remove_tags
    # wasn't invented yet.
    from django.utils.html import remove_tags

    def remove_p_and_br(value):
        return remove_tags(value, 'p br')

except ImportError:
    def remove_p_and_br(value):
        value = re.sub(r'</?p>', '', value)
        value = re.sub(r'<br ?/?>', '', value)
        return value

from django.utils.text import capfirst
from django.utils.encoding import force_text
from django import forms
from django.forms.models import inlineformset_factory, BaseInlineFormSet
from django.core.files.uploadedfile import UploadedFile

from speeches.fields import FromStartIntegerField
from speeches.models import (Speech, Speaker, Section,
                             Recording, RecordingTimestamp, Tag)
from speeches.widgets import AudioFileInput, DatePickerWidget, TimePickerWidget
from speeches.importers.import_popolo import PopoloImporter

logger = logging.getLogger(__name__)


def verbose_name(model, field):
    return capfirst(model._meta.get_field(field).verbose_name)


class CleanAudioMixin(object):
    def clean_audio(self):
        audio = self.cleaned_data['audio']
        if audio and isinstance(audio, UploadedFile):
            ext = os.path.splitext(audio.name)[1]
            # Check that the file is an audio file and one of the
            # filetypes we can use
            if (audio.content_type[0:6] != 'audio/' or
                    ext not in ('.ogg', '.mp3', '.wav', '.3gp')):
                raise forms.ValidationError(
                    'That file does not appear to be an audio file')
        return audio


class SpeechAudioForm(forms.ModelForm, CleanAudioMixin):
    class Meta:
        model = Speech
        fields = ('audio',)
        widgets = {
            'audio': AudioFileInput,
        }


class SectionPickForm(forms.Form):
    section = forms.ModelChoiceField(
        label=_('Assign to section'),
        queryset=Section.objects.all(),
        required=True,
        )


class Select2Widget(AutoHeavySelect2Widget):
    def __init__(self, *args, **kwargs):
        # AutoHeavySelect2Mixin's __init__ hard codes 'data_view' without
        # giving us the chance to add in the namespace we need. Let's
        # hard code it ourselves and then skip past
        # AutoHeavySelect2Mixin in the MRO to HeavySelect2Widget and
        # continue from there.
        kwargs['data_view'] = "speeches:django_select2_central_json"

        HeavySelect2Widget.__init__(self, *args, **kwargs)

    def value_from_datadict(self, data, files, name):
        # Inspiration from MultipleSelect2HiddenInput
        """We need to actually alter the form's self.data so that the form rendering
        works, as that's how the Select2 TagFields function. So always return
        the underlying list from the datadict."""

        # We want the value to always be a list when it's non-empty so that
        # below in the clean method on CreateAutoModelSelect2Field we can
        # change the value in place rather than replacing it.

        # If data is non-trivial and not a MultiValueDict, then an error will
        # be thrown, which is a good thing.
        if data:
            return data.getlist(name)

    def render(self, name, value, attrs=None, choices=()):
        """Because of the above; if we are given a list here, we don't want it."""
        if isinstance(value, list):
            value = value[0] if len(value) else None
        return super(Select2Widget, self).render(name, value, attrs, choices)


class Select2CreateWidget(Select2Widget):
    """This widget is to allow someone to select an existing database entry,
    but also create a new one if needed."""
    def init_options(self):
        super(Select2Widget, self).init_options()
        # Version 4.3 of django-select2 changed, in
        # https://github.com/applegrew/django-select2/commit/36bac057, how
        # functions could be passed to the browser unaltered, using string
        # markers.
        self.options['createSearchChoice'] = '*START*django_select2.createSearchChoice*END*'


class StripWhitespaceField(forms.CharField):
    def clean(self, value):
        value = super(StripWhitespaceField, self).clean(value)
        if value:
            value = value.strip()
        return value


class CreateAutoModelSelect2Field(AutoModelSelect2Field):
    empty_values = [None, '', [], (), {}]

    # If anything tries to run a query on .queryset, it means we've missed
    # somewhere where we needed to limit things to the instance
    queryset = 'UNUSED'

    # instance will be set to an instance in the django-subdomain-instances
    # sense by the form
    instance = None

    widget = Select2CreateWidget

    def __init__(self, *args, **kwargs):
        self.search_fields = ['%s__icontains' % self.column]
        if 'widget' in kwargs or 'empty_label' in kwargs:
            logger.warn(
                'widget/empty_label will be overwritten by using this field')
        kwargs['widget'] = self.widget(
            select2_options={'placeholder': ' ', 'width': '100%'})
        kwargs['empty_label'] = ''
        kwargs['required'] = kwargs.get('required', False)
        return (super(CreateAutoModelSelect2Field, self)
                .__init__(*args, **kwargs))

    def to_python(self, value):
        value = value.strip()

        # Inspiration from HeavyModelSelect2TagField
        if value in self.empty_values:
            return None
        try:
            key = self.to_field_name or 'pk'
            value = self.queryset.get(**{key: value})
        except (ValueError, self.model.DoesNotExist):
            value = self.create_model(value)
        return value

    def create_model(self, value):
        # Note that value could currently arrive as two different things:
        # 1) a stringified integer representing the id of an existing
        #    object, or
        # 2) any other string representing the representative string of
        #    the object.
        value = self.queryset.create(
            **{self.column: value, 'instance': self.instance})
        return value

    def clean(self, value):
        # Inspiration from HeavyModelSelect2TagField
        """We'll get a list here, due to the widget; we'll clean the first item,
        and be sure to alter the list that's been passed in. Because that list
        is used in any future render, and if we don't do it this way we get
        exceptions as it's not an integer..."""
        if value:
            v = value.pop()
            v = super(CreateAutoModelSelect2Field, self).clean(v)
            if v:
                value.append(v.pk)
                value = v
            else:
                value = None
        else:
            value = None

        return value


class NonCreateAutoModelSelect2Field(CreateAutoModelSelect2Field):
    widget = Select2Widget

    def create_model(self, value):
        raise forms.ValidationError(_('You must select an existing speaker'))


class SpeakerField(CreateAutoModelSelect2Field):
    model = Speaker
    column = 'name'


class NonCreateSpeakerField(NonCreateAutoModelSelect2Field):
    model = Speaker
    column = 'name'


class SectionField(CreateAutoModelSelect2Field):
    model = Section
    column = 'heading'


class NonCreateSectionField(NonCreateAutoModelSelect2Field):
    model = Section
    column = 'heading'


class SpeechTextFieldWidget(forms.Textarea):
    def render(self, name, value, attrs=None):
        # These first two steps are also done in the superclass, but it's
        # safe to repeat them here to get value to the right state.
        if value is None:
            value = ''
        value = force_text(value)
        value = re.sub(r'<br ?/?>', '<br />\n', value)
        value = remove_p_and_br(value)

        return super(SpeechTextFieldWidget, self).render(name, value, attrs)


class SpeechTextField(StripWhitespaceField):
    widget = SpeechTextFieldWidget

    def clean(self, value):
        value = super(SpeechTextField, self).clean(value)

        # It there is a value, and it's not already been HTMLified
        # then we want to use linebreaks to give it appropriate newlines.
        if value:
            if value.startswith('<p>'):
                value = re.sub(r'</p>\n*<p>', '</p>\n\n<p>', value)
            else:
                value = linebreaks(value)
        return value


class SpeechForm(forms.ModelForm, CleanAudioMixin):
    audio_filename = forms.CharField(widget=forms.HiddenInput, required=False)
    text = SpeechTextField(required=False, label=verbose_name(Speech, 'text'))
    speaker = SpeakerField(label=verbose_name(Speech, 'speaker'))
    section = SectionField(label=verbose_name(Speech, 'section'))

    type = forms.ChoiceField(
        label=verbose_name(Speech, 'type'),
        choices=Speech._meta.get_field('type').choices,
        widget=forms.HiddenInput, required=False,
        )

    start_date = forms.DateField(
        label=verbose_name(Speech, 'start_date'),
        widget=DatePickerWidget,
        required=False,
        localize=True,
        )
    start_time = forms.TimeField(
        input_formats=['%H:%M', '%H:%M:%S'],
        label=verbose_name(Speech, 'start_time'),
        widget=TimePickerWidget,
        required=False,
        )
    end_date = forms.DateField(
        label=verbose_name(Speech, 'end_date'),
        widget=DatePickerWidget,
        required=False,
        localize=True,
        )
    end_time = forms.TimeField(
        input_formats=['%H:%M', '%H:%M:%S'],
        label=verbose_name(Speech, 'end_time'),
        widget=TimePickerWidget,
        required=False,
        )
    # tags = TagField()
    tags = forms.ModelMultipleChoiceField(
        queryset=Tag.objects.all(),
        label=verbose_name(Speech, 'tags'),
        widget=Select2MultipleWidget(
            select2_options={'placeholder': _('Choose tags'),
                             'width': 'resolve'}),
        required=False,
        )

    def clean(self):
        cleaned_data = self.cleaned_data

        # Temporary fix - see issue #340 (
        if not cleaned_data.get('type'):
            cleaned_data['type'] = ('speech' if cleaned_data.get('speaker')
                                    else 'narrative')

        if 'audio_filename' in cleaned_data and cleaned_data['audio_filename']:
            filename = cleaned_data['audio_filename']
            self.cleaned_data['audio'] = filename

        if not cleaned_data.get('text') and not cleaned_data.get('audio'):
            raise forms.ValidationError(
                _('You must provide either text or some audio'))

        # If we have text but no speaker, then this should become a
        # <narrative> element in the Akoma Ntoso, which can't contain
        # <p> elements, so we should replace any in the middle with
        # <br /> and get rid of the ones round the outside.
        if 'text' in cleaned_data and not cleaned_data.get('speaker'):
            text = cleaned_data['text']
            text = re.sub(r'</p>\n\n<p>', '<br />\n', text)
            text = re.sub(r'</?p>', '', text)
            cleaned_data['text'] = text

        return cleaned_data

    def clean_start_time(self):
        if (self.cleaned_data['start_time'] and
                not self.cleaned_data.get('start_date')):
            raise forms.ValidationError(
                _('If you provide a start time you must give a start date too')
                )
        return self.cleaned_data['start_time']

    def clean_end_time(self):
        if self.cleaned_data['end_time'] and not self.cleaned_data['end_date']:
            raise forms.ValidationError(
                _('If you provide an end time you must give an end date too'))
        return self.cleaned_data['end_time']

    class Meta:
        model = Speech
        widgets = {
            'audio': AudioFileInput,
            'event': forms.TextInput(),
            'heading': forms.TextInput(),
            'location': forms.TextInput(),
            'source_url': forms.TextInput(),
        }
        exclude = ('celery_task_id', 'instance')


class RecordingAPIForm(forms.ModelForm, CleanAudioMixin):
    # Form for uploading a recording

    # Force timestamps to be a charfield so we can supply json to it
    timestamps = forms.CharField(required=False)

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request')
        return super(RecordingAPIForm, self).__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = self.cleaned_data
        if 'audio_filename' in cleaned_data and cleaned_data['audio_filename']:
            filename = cleaned_data['audio_filename']
            self.cleaned_data['audio'] = filename
        return cleaned_data

    def clean_timestamps(self):
        # Allow uploading of timestamps which don't exist by turning
        # the supplied dictionary of timestamps into new
        # RecordingTimestamp instances
        timestamps = []

        if ('timestamps' not in self.cleaned_data or
                not self.cleaned_data['timestamps']):
            logger.debug('No timestamps in cleaned_data')
            return timestamps

        logger.debug(
            'timestamps received = %s' % self.cleaned_data['timestamps'])
        timestamps_json = json.loads(self.cleaned_data['timestamps'])

        if not isinstance(timestamps_json, list):
            return timestamps

        for recording_timestamp in timestamps_json:
            try:
                if 'timestamp' in recording_timestamp:
                    # Note - we divide by 1000 because the time comes
                    # from javascript and is in milliseconds, but this
                    # expects the time in seconds
                    supplied_time = int(recording_timestamp['timestamp'] / 1000)
                    # We also make it a UTC time!
                    timestamp = (datetime.utcfromtimestamp(supplied_time)
                                 .replace(tzinfo=pytz.utc))
                    try:
                        speaker = Speaker.objects.get(
                            pk=recording_timestamp['speaker'])
                    except:
                        speaker = None
                    timestamps.append(
                        RecordingTimestamp(
                            speaker=speaker,
                            timestamp=timestamp,
                            instance=self.request.instance)
                        )
                else:
                    # Timestamp is required
                    logger.error(
                        "No timestamp supplied in request: {0}".format(
                            recording_timestamp)
                        )
            except ValueError:
                # Ignore this one
                logger.error(
                    "ValueError encountered parsing: {0} into speaker/timestamp"
                    .format(recording_timestamp)
                    )

        if timestamps.count == 0:
            logger.error(
                "Timestamps parameter was given but no timestamps parsed from: {0}"
                .format(self.cleaned_data['timestamp'])
                )

        return timestamps

    class Meta:
        model = Recording
        exclude = ('instance',)


class RecordingForm(forms.ModelForm):
    class Meta:
        model = Recording
        exclude = ['instance', 'audio']


class SectionForm(forms.ModelForm):
    heading = forms.CharField(
        required=True,
        label=verbose_name(Section, 'heading'),
        )
    parent = NonCreateSectionField(label=verbose_name(Section, 'parent'))
    start_date = forms.DateField(
        label=verbose_name(Section, 'start_date'),
        widget=DatePickerWidget,
        required=False,
        localize=True,
        )
    start_time = forms.TimeField(
        input_formats=['%H:%M', '%H:%M:%S'],
        label=verbose_name(Section, 'start_time'),
        widget=TimePickerWidget,
        required=False,
        )
    source_url = forms.CharField(label=verbose_name(Section, 'source_url'), required=False)

    def __init__(self, *args, **kwargs):
        super(SectionForm, self).__init__(*args, **kwargs)
        if self.instance.id:
            ids = [self.instance.id]
            ids.extend([d.id for d in self.instance.get_descendants])
            self.fields['parent'].queryset = Section.objects.exclude(id__in=ids)

    class Meta:
        model = Section
        fields = ('heading', 'description', 'parent', 'source_url', 'start_date', 'start_time')

    def clean_parent(self):
        parent = self.cleaned_data['parent']
        if self.instance and parent:
            if parent.id == self.instance.id:
                raise forms.ValidationError(
                    _('Something cannot be its own parent'))
            descendant_ids = [d.id for d in self.instance.get_descendants]
            if parent.id in descendant_ids:
                raise forms.ValidationError(
                    _('Something cannot have a parent that is also a descendant')
                    )
        return parent


class SpeakerForm(forms.ModelForm):
    name = StripWhitespaceField(label=verbose_name(Speaker, 'name'))

    class Meta:
        model = Speaker
        fields = ('name', 'image', 'summary', 'sort_name')


class SpeakerDeleteForm(forms.ModelForm):
    new_speaker = NonCreateSpeakerField(
        label=_('New speaker'), required=False)

    def __init__(self, *args, **kwargs):
        super(SpeakerDeleteForm, self).__init__(*args, **kwargs)

        count = self.instance.speech_set.count()

        label = ungettext(
            "This speaker has %(count)d speech. What would you like to do with it?",
            "This speaker has %(count)d speeches. What would you like to do with them?",
            count
            ) % {'count': count}

        choices = (
            ('Reassign', ungettext('Assign it to another speaker', 'Assign them to another speaker', count)),
            ('Narrative', ungettext(
                'Make it into narrative (i.e. remove the speaker)',
                'Make them into narrative (i.e. remove the speaker)', count)),
            ('Delete', ungettext('Delete it', 'Delete them', count)),
            )

        self.fields['action'] = forms.ChoiceField(
            label=label,
            choices=choices,
            widget=forms.RadioSelect(),
            )

    def clean(self):
        super(SpeakerDeleteForm, self).clean()
        new_speaker = self.cleaned_data.get('new_speaker')

        if self.cleaned_data.get('action') == 'Reassign':
            if new_speaker and new_speaker.id == self.instance.id:
                self._errors['new_speaker'] = self.error_class(
                    [_("You can't assign speeches to the speaker you're deleting")]
                    )
            if not new_speaker and 'new_speaker' not in self._errors:
                # When Django 1.7 is our oldest supported version, we can use add_error
                # https://docs.djangoproject.com/en/1.7/ref/forms/api/#django.forms.Form.add_error
                self._errors['new_speaker'] = self.error_class(
                    [_('You must choose a new speaker if reassigning speeches')]
                    )
        else:
            if new_speaker:
                # As above when our oldest supported version of Django is 1.7
                self._errors['new_speaker'] = self.error_class(
                    [_('You must not choose a new speaker unless reassigning speeches')]
                    )
        return self.cleaned_data

    class Meta:
        model = Speaker
        fields = ('id',)


class RecordingTimestampForm(forms.ModelForm):
    timestamp = FromStartIntegerField()

    def __init__(self, *args, **kwargs):
        super(RecordingTimestampForm, self).__init__(*args, **kwargs)
        # Each timestamp needs to know the recording start time if it
        # is bound (e.g. not an extra, blank field)
        if self.instance.recording_id:
            self.fields['timestamp'].recording_start = self.instance.recording.start_datetime

    def save(self, commit=True):
        self.instance.instance_id = self.instance.recording.instance_id
        return super(RecordingTimestampForm, self).save(commit)

    class Meta:
        model = RecordingTimestamp
        exclude = ['instance', 'speech']


class BaseRecordingTimestampFormSet(BaseInlineFormSet):
    def clean(self):
        if any(self.errors):
            return

        recording = self.instance

        # we're using '_forms' to avoid clashing with forms import, e.g.
        # for ValidationError
        _forms = sorted(
            [f for f in self.forms if 'timestamp' in f.cleaned_data],
            key=lambda f: f.cleaned_data['timestamp']
        )

        first_timestamp = _forms[0].cleaned_data['timestamp']
        last_timestamp = _forms[-1].cleaned_data['timestamp']

        # TODO: check that first timestamp isn't before start of speech?

        if first_timestamp < recording.start_datetime:
            raise forms.ValidationError(
                _("Start time is before recording start time!"))

        # TODO: check that delta from first to last timestamp isn't longer
        # than length of audio
        # This is slightly complicated because we don't seem to cache this
        # metadata anywhere?  Might make sense to add to Recording?

        delta = (last_timestamp - first_timestamp).seconds
        if delta >= recording.audio_duration:
            raise forms.ValidationError(
                _('Difference between timestamps is too long for the uploaded audio'))

        previous_timestamp = None
        for form in _forms:
            timestamp = form.cleaned_data['timestamp']
            if previous_timestamp:
                if timestamp <= previous_timestamp:
                    raise forms.ValidationError(
                        _('Timestamps must be distinct'))
            previous_timestamp = timestamp

RecordingTimestampFormSet = inlineformset_factory(
    Recording,
    RecordingTimestamp,
    formset=BaseRecordingTimestampFormSet,
    form=RecordingTimestampForm,
    extra=1,
    can_delete=1,
)


class PopoloImportForm(forms.Form):
    location = forms.URLField(
        label=_('Location of Popolo JSON data'))

    def __init__(self, instance=None, *args, **kwargs):
        super(PopoloImportForm, self).__init__(*args, **kwargs)
        self.instance = instance

    def clean(self):
        cleaned_data = super(PopoloImportForm, self).clean()

        cleaned_data['importer'] = PopoloImporter(
            cleaned_data['location'],
            instance=self.instance,
            )

        return cleaned_data


class AkomaNtosoImportForm(forms.Form):
    location = forms.URLField(label=_('Location of Akoma Ntoso data'))

    def __init__(self, instance=None, *args, **kwargs):
        super(AkomaNtosoImportForm, self).__init__(*args, **kwargs)
        self.instance = instance

        if Section.objects.filter(instance=instance).exists():
            self.fields['existing_sections'] = forms.ChoiceField(
                label=_('What would you like to do with existing top level sections?'),
                choices=(
                    ('skip', _('Skip them - keep them exactly as they are')),
                    ('replace', _('Replace them - throw away the existing data and use the new')),
                    ('merge', _('Merge the new data into the existing sections - things in both will be duplicated')),
                    ),
                widget=forms.RadioSelect(),
                )



import datetime
import json
import re
import requests
from mock import patch

from speeches.tests import InstanceTestCase
from speeches.models import Speech, Speaker, Section
from speeches.importers.import_akomantoso import ImportAkomaNtoso
from speeches.importers.import_popolo import PopoloImporter


class FakeRequestsOutput(object):
    def __init__(self, source, verify=True):
        assert source.startswith('http://example.com/')

        # We'll put things that would have been served from a url ending
        # in a / at the same place ending in and _, so as to avoid a
        # name clash with directories.
        source = re.sub(r'/$', '_', source)
        source = re.sub(
            r'^http://example.com/',
            'speeches/tests/data/fake_http/',
            source)

        self.file_path = source

    def json(self):
        return json.load(open(self.file_path))

    @property
    def content(self):
        return open(self.file_path, 'rb').read()


@patch.object(requests, 'get', FakeRequestsOutput)
class AkomaNtosoImportTestCase(InstanceTestCase):
    def setUp(self):
        super(AkomaNtosoImportTestCase, self).setUp()
        self.importer = ImportAkomaNtoso(instance=self.instance, commit=True)

    def _list_sections(self):
        # Make mapping {section name: list of its speeches' texts}
        sections = {section.id: [] for section in Section.objects.all()}
        for speech in Speech.objects.all():
            if speech.section_id:
                sections[speech.section_id].append(speech.text)
        return {section.title: sections[section.id] for section in Section.objects.all()}

    def test_import_sample_file(self):
        self.importer.import_document(
            'speeches/tests/data/fake_http/Debate_Bungeni_1995-10-31.xml')

        # To get us started, let's just check that we get the right kind of
        # speech in the right order.
        self.assertEqual(
            [x.type for x in Speech.objects.all()],
            [u'scene', u'other', u'narrative', u'speech', u'question',
             u'summary', u'speech', u'answer', u'narrative', u'speech',
             u'narrative']
            )

    def test_already_imported(self):
        self.importer.import_document(
            'speeches/fixtures/test_inputs/test_xpath.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Hello</p>']}
            )

        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='skip').import_document(
            'speeches/tests/data/fake_http/test_clobber.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Hello</p>']}
            )

        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='merge').import_document(
            'speeches/tests/data/fake_http/test_clobber.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Hello</p>', '<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })

        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='replace').import_document(
            'speeches/tests/data/fake_http/test_clobber.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })

        ImportAkomaNtoso(instance=self.instance, commit=True).import_document(
            'speeches/tests/data/fake_http/test_clobber.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             'This is the title': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })

    def test_not_already_imported(self):
        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='skip').import_document(
            'speeches/tests/data/fake_http/test_clobber.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })
        Section.objects.all().delete()

        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='merge').import_document(
            'speeches/tests/data/fake_http/test_clobber.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })
        Section.objects.all().delete()

        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='replace').import_document(
            'speeches/tests/data/fake_http/test_clobber.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })
        Section.objects.all().delete()

        ImportAkomaNtoso(instance=self.instance, commit=True).import_document(
            'speeches/tests/data/fake_http/test_clobber.xml')
        self.assertEqual(
            self._list_sections(),
            {'This is the title': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })
        Section.objects.all().delete()

    def test_empty_title(self):
        self.importer.import_document(
            'speeches/fixtures/test_inputs/test_empty_title.xml')
        self.assertEqual(
            self._list_sections(),
            {'Untitled': ['<p>Hello</p>'],
             'Untitled': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })

        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='skip').import_document(
            'speeches/fixtures/test_inputs/test_empty_title.xml')
        self.assertEqual(
            self._list_sections(),
            {'Untitled': ['<p>Hello</p>'],
             'Untitled': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })

        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='merge').import_document(
            'speeches/fixtures/test_inputs/test_empty_title.xml')
        self.assertEqual(
            self._list_sections(),
            {'Untitled': ['<p>Hello</p>', '<p>Hello</p>', '<p>Howdy</p>'],
             'Untitled': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>', '<p>Bye</p>'],
             })

        ImportAkomaNtoso(instance=self.instance, commit=True, clobber='replace').import_document(
            'speeches/fixtures/test_inputs/test_empty_title.xml')
        self.assertEqual(
            self._list_sections(),
            {'Untitled': ['<p>Hello</p>'],
             'Untitled': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })

        ImportAkomaNtoso(instance=self.instance, commit=True).import_document(
            'speeches/fixtures/test_inputs/test_empty_title.xml')
        self.assertEqual(
            self._list_sections(),
            {'Untitled': ['<p>Hello</p>'],
             'Untitled': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             'Untitled': ['<p>Hello</p>'],
             'Untitled': ['<p>Howdy</p>'],
             'Conclusions': ['<p>Bye</p>'],
             })

    def test_empty_docDate(self):
        self.importer.import_document(
            'speeches/fixtures/test_inputs/test_empty_docDate.xml')
        self.assertEqual(
            [(x.start_date, x.title, x.source_url) for x in Section.objects.all()],
            [(datetime.date(2012, 3, 7), 'Title', 'http://example.org')]
        )

    def test_xpath_preface_elements(self):
        self.importer.import_document(
            'speeches/fixtures/test_inputs/test_xpath.xml')
        self.assertEqual(
            [x.title for x in Section.objects.all()],
            ['This is the title']
        )
        self.assertEqual(
            [x.start_date for x in Speech.objects.all()],
            [datetime.date(2014, 7, 24)]
        )

    def test_unicode_character(self):
        self.importer.import_document(
            'speeches/fixtures/test_inputs/test_unicode_character.xml')

        self.assertEqual(
            [x.type for x in Speech.objects.all()],
            ['other']
            )

    def test_blank_speakers(self):
        self.importer.import_document(
            'speeches/fixtures/test_inputs/test_blank_speakers.xml')

        speaker = Speaker.objects.get(name='Speaker')
        speeches = Speech.objects.all()
        speeches_s = Speech.objects.filter(type='speech')
        self.assertEqual(speeches.count(), speeches_s.count())

        for i in range(4):
            s = speaker if i % 2 else None
            sd = 'Speaker' if i > 1 else None
            self.assertEqual(speeches[i].speaker, s)
            self.assertEqual(speeches[i].speaker_display, sd)

    def test_import_remote_file(self):
        self.importer.import_document(
            'http://example.com/Debate_Bungeni_1995-10-31.xml')

        # To get us started, let's just check that we get the right kind of
        # speech in the right order.
        self.assertEqual(
            [x.type for x in Speech.objects.all()],
            [u'scene', u'other', u'narrative', u'speech', u'question',
             u'summary', u'speech', u'answer', u'narrative', u'speech',
             u'narrative']
            )


@patch.object(requests, 'get', FakeRequestsOutput)
class AkomaNtosoImportViewTestCase(InstanceTestCase):
    def test_import_page_smoke_test(self):
        resp = self.client.get('/import/akomantoso')

        self.assertContains(resp, 'Import speeches')

    def test_import_data(self):
        resp = self.client.post(
            '/import/akomantoso',
            {'location': 'http://example.com/Debate_Bungeni_1995-10-31.xml'},
            follow=True,
            )

        # To get us started, let's just check that we get the right kind of
        # speech in the right order.
        self.assertEqual(
            [x.type for x in Speech.objects.all()],
            [u'scene', u'other', u'narrative', u'speech', u'question',
             u'summary', u'speech', u'answer', u'narrative', u'speech',
             u'narrative']
            )

        self.assertContains(resp, 'Created: 7 speakers, 8 sections, 11 speeches')

    def test_import_data_twice_skipped(self):
        for i in range(2):
            resp = self.client.post(
                '/import/akomantoso',
                {'location': 'http://example.com/test_clobber.xml', 'existing_sections': 'skip'},
                follow=True,
                )

        self.assertEqual(
            [x.type for x in Speech.objects.all()],
            [u'speech', u'speech']
            )

        self.assertContains(resp, 'Nothing new to import.')

    def test_import_bad_data(self):
        resp = self.client.post(
            '/import/akomantoso',
            {'location': 'http://example.com/welsh_assembly/persons'},
            )

        self.assertContains(resp, 'Sorry - something went wrong with the import')


@patch.object(requests, 'get', FakeRequestsOutput)
class PopitImportTestCase(InstanceTestCase):
    popit_url = 'http://example.com/welsh_assembly_popit/'

    def test_popit_import_persons(self):
        popit_importer = PopoloImporter(self.popit_url)
        popit_importer.import_all()

        self.assertEqual(
            Speaker.objects.filter(instance=popit_importer.instance).count(),
            3,
            )


@patch.object(requests, 'get', FakeRequestsOutput)
class PopoloImportTestCase(InstanceTestCase):
    popit_url = 'http://example.com/welsh_assembly/persons'

    def test_popit_import_persons(self):
        popolo_importer = PopoloImporter(self.popit_url)
        popolo_importer.import_all()

        self.assertEqual(Speaker.objects.filter(instance=popolo_importer.instance).count(), 3)

    def test_popolo_import_remote_single_json_file(self):
        popolo_importer = PopoloImporter(
            'http://example.com/welsh_assembly.json'
            )
        popolo_importer.import_all()

        self.assertEqual(
            Speaker.objects.filter(instance=popolo_importer.instance).count(),
            3,
            )


class PopoloImportFromLocalSourceTestCase(InstanceTestCase):
    def test_import_persons(self):
        popolo_importer = PopoloImporter(
            'speeches/tests/data/fake_http/welsh_assembly.json'
            )
        popolo_importer.import_all()

        self.assertEqual(
            Speaker.objects.filter(instance=popolo_importer.instance).count(),
            3,
            )


@patch.object(requests, 'get', FakeRequestsOutput)
class PopoloImportViewsTestCase(InstanceTestCase):
    def test_import_page_smoke_test(self):
        resp = self.client.get('/import/popolo')

        self.assertContains(resp, 'Import speakers')

    def test_import_with_data(self):
        resp = self.client.post(
            '/import/popolo',
            {'location': 'http://example.com/welsh_assembly/persons'},
            follow=True,
            )

        self.assertEqual(
            Speaker.objects.filter(instance=self.instance).count(),
            3,
            )
        self.assertContains(resp, '3 speakers created. 0 speakers refreshed.')

        # Repeat the same post
        resp = self.client.post(
            '/import/popolo',
            {'location': 'http://example.com/welsh_assembly/persons'},
            follow=True,
            )

        self.assertEqual(
            Speaker.objects.filter(instance=self.instance).count(),
            3,
            )
        self.assertContains(resp, '0 speakers created. 3 speakers refreshed.')

    def test_import_empty(self):
        resp = self.client.post(
            '/import/popolo',
            {'location': 'http://example.com/empty.json'},
            follow=True,
            )

        self.assertEqual(
            Speaker.objects.filter(instance=self.instance).count(),
            0,
            )
        self.assertContains(resp, 'No speakers found')



import json
import unittest

from speeches.tests import InstanceTestCase
from speeches.models import Speech

try:
    import tastypie  # noqa
    skip_tastypie = False
except ImportError:
    skip_tastypie = True


@unittest.skipIf(skip_tastypie, 'Tastypie tests skipping, as not installed')
class SpeechAPITests(InstanceTestCase):

    def setUp(self):
        super(SpeechAPITests, self).setUp()

        # create the public and private speeches
        self.public_speech = Speech.objects.create(heading="Public", public=True, instance=self.instance)
        self.private_speech = Speech.objects.create(heading="Private", public=False, instance=self.instance)

    def get_speech_from_api(self, speech):
        url = '/api/v0.1/speech/' + str(speech.id) + '/?format=json'
        # print "%s (%s): %s" % (speech, speech.id, url)
        return self.client.get(url)

    def test_speech_list(self):
        resp = self.client.get('/api/v0.1/speech/?format=json')
        data = json.loads(resp.content.decode())
        objects = data['objects']

        self.assertEqual(len(objects), 1)

        self.assertEqual(objects[0]['heading'], "Public")

    def test_public_speech_visible(self):
        resp = self.get_speech_from_api(self.public_speech)
        self.assertEqual(resp.status_code, 200)

    def test_private_speech_hidden(self):
        resp = self.get_speech_from_api(self.private_speech)
        self.assertEqual(resp.status_code, 404)



from optparse import make_option
from six.moves.urllib.request import urlopen
import xml.etree.ElementTree as etree

from django.core.management.base import BaseCommand, CommandError

from speeches.models import Section, Speech, Speaker
from instances.models import Instance

PLAYS = {
    "all_well.xml": "All's Well That Ends Well",
    "as_you.xml": "As You Like It",
    "a_and_c.xml": "Antony and Cleopatra",
    "com_err.xml": "A Comedy of Errors",
    "coriolan.xml": "Coriolanus",
    "cymbelin.xml": "Cymbeline",
    "dream.xml": "A Midsummer Night's Dream",
    "hamlet.xml": "Hamlet",
    "hen_iv_1.xml": "Henry IV, Part I",
    "hen_iv_2.xml": "Henry IV, Part II",
    "hen_v.xml": "Henry V",
    "hen_viii.xml": "Henry VIII",
    "hen_vi_1.xml": "Henry VI, Part 1",
    "hen_vi_2.xml": "Henry VI, Part 2",
    "hen_vi_3.xml": "Henry VI, Part 3",
    "john.xml": "The Life and Death of King John",
    "j_caesar.xml": "Julius Caesar",
    "lear.xml": "King Lear",
    "lll.xml": "Love's Labor's Lost",
    "macbeth.xml": "Macbeth",
    "merchant.xml": "The Merchant of Venice",
    "much_ado.xml": "Much Ado About Nothing",
    "m_for_m.xml": "Measure for Measure",
    "m_wives.xml": "The Merry Wives of Windsor",
    "othello.xml": "Othello",
    "pericles.xml": "Pericles",
    "rich_ii.xml": "Richard II",
    "rich_iii.xml": "Richard III",
    "r_and_j.xml": "Romeo and Juliet",
    "taming.xml": "The Taming of the Shrew",
    "tempest.xml": "The Tempest",
    "timon.xml": "Timon of Athens",
    "titus.xml": "Titus Andronicus",
    "troilus.xml": "Troilus and Cressida",
    "two_gent.xml": "Two Gentlemen of Verona",
    "t_night.xml": "Twelfth Night",
    "win_tale.xml": "A Winter's Tale",
}


class Command(BaseCommand):
    args = '<play>'
    help = 'Import a Shakespeare play into a SayIt instance'
    option_list = BaseCommand.option_list + (
        make_option('--commit', action='store_true', help='Whether to commit to the database or not'),
        make_option('--instance', action='store', default='shakespeare', help='Label of instance to add data to'),
        make_option('--list', action='store_true', help='List the plays available'),
    )

    def make(self, cls, **kwargs):
        s = cls(instance=self.instance, **kwargs)
        if self.commit:
            s.save()
        elif s.heading:
            print(s.heading)
        return s

    def handle(self, *args, **options):
        if options['list'] or len(args) != 1:
            self.stdout.write('Plays:\n')
            for play in sorted(PLAYS.values()):
                self.stdout.write('* %s\n' % play)
            if not options['list']:
                raise CommandError("Please specify a play")
            return

        play = args[0]
        file = None
        for f, p in PLAYS.items():
            if play == p:
                file = f
                break

        if not file:
            raise CommandError("No matching play found")

        try:
            self.instance = Instance.objects.get(label=options['instance'])
        except:
            raise CommandError("Instance specified not found")

        self.commit = options['commit']

        xml = urlopen('http://www.ibiblio.org/xml/examples/shakespeare/%s' % file).read()
        play_xml = etree.fromstring(xml)
        play_section = self.make(Section, heading=play)

        speakers = {}
        for act in play_xml:
            if act.tag != 'ACT':
                continue
            act_heading = act[0].text
            act_section = self.make(Section, heading=act_heading, parent=play_section)
            scenes = act[1:]
            for scene in scenes:
                scene_heading = scene[0].text
                scene_section = self.make(Section, heading=scene_heading, parent=act_section)
                speeches_xml = scene[1:]
                for sp in speeches_xml:
                    if sp.tag == 'STAGEDIR' or sp.tag == 'SUBHEAD' or sp.tag == 'SUBTITLE':
                        self.make(Speech, section=scene_section, text='<p><i>%s</i></p>' % sp.text, type='narrative')
                        continue

                    if not sp[0].text:
                        speaker = None
                    elif self.commit:
                        name = sp[0].text.replace('[', '').replace(']', '')
                        if name in speakers:
                            speaker = speakers[name]
                        else:
                            speaker = Speaker.objects.create(name=name, instance=self.instance)
                            speakers[name] = speaker
                    else:
                        speaker = Speaker(name=sp[0].text, instance=self.instance)

                    text = ""
                    lines = sp[1:]
                    for line in lines:
                        if len(line):
                            text += '<i>%s</i>' % line[0].text
                            if line[0].tail:
                                text += ' %s' % line[0].tail.strip()
                            text += '<br>\n'
                        elif line.tag == 'LINE':
                            text += '%s<br>\n' % line.text
                        elif line.tag == 'STAGEDIR':
                            text += '<i>%s</i><br>\n' % line.text

                    text = '<p>%s</p>' % text
                    self.make(Speech, speaker=speaker, section=scene_section, text=text, type='speech')



import random

from django.test import TestCase

from speeches.models import cache


class CacheTest(object):
    @cache
    def random(self):
        # return a different value each time, if not cached
        return random.random()


class CacheTests(TestCase):

    def test_caching(self):

        # Create an object, ensure that the caching occurs.
        cachetest1 = CacheTest()
        rand1 = cachetest1.random
        rand2 = cachetest1.random
        self.assertEqual(rand1, rand2)

        # Create another object, check new value cached.
        cachetest2 = CacheTest()
        rand3 = cachetest2.random
        self.assertNotEqual(rand1, rand3)



import os
import subprocess
import logging
import tempfile
import calendar
import audioread.ffdec

logger = logging.getLogger(__name__)


class AudioException(Exception):
    """Custom exception for audio transcoding ThingsThatGoWrong"""
    pass


class AudioHelper(object):

    def make_wav(self, in_filename):
        """Make a .wav file suitable for uploading to AT&T and return true if
           it succeeded"""
        # First make a temporary file
        (fd, out_filename) = tempfile.mkstemp(suffix='.wav')

        options = self._build_ffmpeg_options(in_filename)
        options.extend([
            # Output options
            # Sample rate of 8KHz
            '-ar',
            '8000',
            # Single channel, ie: mono
            '-ac',
            '1',
            # Use the 16-bit pcm codec
            '-acodec',
            'pcm_s16le',
            # Output file
            out_filename

        ])

        with open(os.devnull, 'w') as dev_null:
            result = subprocess.call(options, stderr=dev_null)

        if not result == 0:
            message = "FFMPEG failed to make .wav with result: {0} on file: {1}".format(result, in_filename)
            logger.error(message)
            raise AudioException(message)

        return out_filename

    def make_mp3(self, in_filename):
        """Make a .mp3 file suitable for displaying publically on the website, returns
           the filename if it succeeded, and throws an AudioException if something went wrong"""
        (fd, out_filename) = tempfile.mkstemp(suffix=".mp3")
        options = self._build_ffmpeg_options(in_filename)
        options.extend(self._build_ffmpeg_mp3_output_options(out_filename))
        with open(os.devnull, 'w') as dev_null:
            result = subprocess.call(options, stderr=dev_null)
        if not result == 0:
            message = "FFMPEG failed to make .mp3 with result: {0} on file: {1}".format(result, in_filename)
            logger.error(message)
            raise AudioException(message)

        return out_filename

    def split_recording(self, recording):
        """Make a series of .mp3 files from one recording, based on its timestamps"""

        # List of files we'll return
        files = []

        # Does it have any audio?
        if not recording.audio:
            logger.error("Asked to split recording: {0} with no audio, returning immediately".format(recording))
            return files

        # Do we have any timestamps to split it by?
        if not recording.timestamps.count():
            # No timestamps, so just make an mp3 of the whole thing
            # Though 1 timestamp also creates just a single speech, that
            # timestamp may be offset, so not included here!
            logger.info("No timestamps in the recording: {0}, so making an mp3 of all of it.".format(recording))
            files.append(self.make_mp3(recording.audio.path))
            return files

        # We have more than one timestamp
        sorted_timestamps = recording.timestamps.all().order_by("timestamp")

        # munge start_timestamp into the format that make_partial_mp3 expects
        start_timestamp = calendar.timegm(recording.start_datetime.timetuple())
        for index, timestamp in enumerate(sorted_timestamps):
            next_timestamp = None
            if index < (len(sorted_timestamps) - 1):
                next_timestamp = sorted_timestamps[index + 1]

            files.append(self.make_partial_mp3(recording, start_timestamp, timestamp, next_timestamp))

        return files

    def make_partial_mp3(self, recording, start_timestamp, timestamp, next_timestamp):
        """Make a partial mp3 file for a given timestamp"""
        # We assume the first timestamp is 00:00:00 in the recording
        # so we can then calculate the other offsets from there.

        # Work out the time this timestamp starts, in seconds from the start of the recording
        start_time_relative = timestamp.utc - start_timestamp

        # Work out the time this timestamp ends, in seconds from the start
        duration = None
        if next_timestamp is not None:
            duration = next_timestamp.utc - timestamp.utc

        (fd, out_filename) = tempfile.mkstemp(suffix=".mp3")

        options = self._build_ffmpeg_options(recording.audio.path)
        # Where to start from
        options.extend(["-ss", str(start_time_relative)])
        # Where to go to
        if duration is not None:
            options.extend(['-t', str(duration)])

        # because the partial files are built from an mp3, we can use the 'copy' pseudocodec
        options.extend(self._build_ffmpeg_mp3_output_options(out_filename, 'copy'))

        with open(os.devnull, 'w') as dev_null:
            result = subprocess.call(options, stderr=dev_null)

        # Check that the result was ok, and if not, blow up
        if not result == 0:
            message = "FFMPEG failed with result: {0} on timestamp: {1}".format(result, timestamp)
            logger.error(message)
            raise AudioException(message)

        return out_filename

    def get_audio_duration(self, in_filename):
        return round(audioread.ffdec.FFmpegAudioFile(in_filename).duration)

    def _build_ffmpeg_options(self, in_filename):
        return [
            'ffmpeg',
            # Tell ffmpeg to shut up
            '-loglevel',
            '0',
            # Say yes to everything
            '-y',
            # Input file
            '-i',
            in_filename
        ]

    def _build_ffmpeg_mp3_output_options(self, out_filename, encoder='libmp3lame'):
        return [
            # Output options
            # 128 kbps bitrate
            '-ab',
            '128k',
            '-acodec',
            encoder,
            # Output file
            out_filename
        ]



# -*- coding: utf-8 -*-

import logging

from dateutil import parser as dateutil
from lxml import etree
from lxml import objectify
import requests

from speeches.importers.import_base import ImporterBase
from speeches.models import Section, Speech, Speaker

logger = logging.getLogger(__name__)


class ImportAkomaNtoso(ImporterBase):
    start_date = None

    def import_document(self, document_path):
        if document_path.startswith('http'):
            self.xml = objectify.fromstring(requests.get(document_path, verify=self.verify).content)
        else:
            self.xml = objectify.parse(document_path).getroot()
        self.ns = self.xml.nsmap.get(None, None)
        return self.parse_document()

    def parse_document(self):
        self.stats = {Speaker: 0}
        debate = self.xml.debate

        if self.ns:
            people = debate.findall(
                'an:meta/an:references/an:TLCPerson',
                namespaces={'an': self.ns},
                )
        else:
            people = debate.findall('meta/references/TLCPerson')
        if people is None:
            people = []
        for person in people:
            id = person.get('id')
            href = person.get('href')
            try:
                speaker = Speaker.objects.get(
                    instance=self.instance, identifiers__identifier=href)
            except Speaker.DoesNotExist:
                name = person.get('showAs')
                if not name:
                    raise Exception("TLCPerson '%s' is missing showAs" % href)
                speaker = Speaker(instance=self.instance, name=name)
                self.stats[Speaker] += 1

                if self.commit:
                    speaker.save()
                    speaker.identifiers.create(
                        identifier=href, scheme='Akoma Ntoso import')

            self.speakers[id] = speaker

        docDate = self.get_preface_tag(debate, 'docDate')
        if docDate is not None:
            date = docDate.get('date')
            if date:
                try:
                    self.start_date = dateutil.parse(date)
                except ValueError:
                    logger.warn("docDate element did not parse '%s'" % date)
            else:
                logger.warn("docDate element missing required date attribute")

        docTitle = self.get_preface_tag(debate, 'docTitle')
        if docTitle:
            docTitle = docTitle.text

        docNumber = self.get_preface_tag(debate, 'docNumber')
        if docNumber:
            docNumber = docNumber.text

        legislature = self.get_preface_tag(debate, 'legislature')
        if legislature:
            legislature = legislature.text

        session = self.get_preface_tag(debate, 'session')
        if session:
            session = session.text

        source_url = self.get_preface_tag(debate, 'link')
        if source_url is not None:
            source_url = source_url.get('href')

        self.imported_section_ids = set()

        section = None
        if docTitle:
            kwargs = {
                'parent': None,
                'heading': docTitle,
                'start_date': self.start_date,
                'number': docNumber or '',
                'legislature': legislature or '',
                'session': session or '',
            }

            section = self.make_section(source_url=source_url or '', **kwargs)

            if not section:
                return self.stats

        self.visit(debate.debateBody, section)
        return self.stats

    def get_preface_tag(self, debate, tag):
        if self.ns:
            tag = debate.xpath('an:coverPage//an:%s|an:preface//an:%s' % (tag, tag), namespaces={'an': self.ns})
        else:
            tag = debate.xpath('coverPage//%s|preface//%s' % (tag, tag))
        if tag:
            return tag[0]

    def make_section(self, source_url='', **kwargs):
        # If the importer has no opinion on clobbering, just import the section,
        # potentially creating a duplicate section.
        if self.clobber:
            qs = Section.objects.for_instance(self.instance).filter(**kwargs)
            if qs:
                if self.clobber == 'replace':
                    logger.info('Replacing %s' % kwargs.get('heading'))
                    # Delete old sections, unless they are from this import
                    for section in qs:
                        if section.id in self.imported_section_ids:
                            break
                        for speech in section.descendant_speeches():
                            speech.delete()
                        section.delete()
                elif self.clobber == 'merge':
                    # Return (any of) existing section(s), unless it is from this import
                    section = qs[0]
                    if section.id in self.imported_section_ids:
                        logger.info('Importing %s' % kwargs.get('heading'))
                    else:
                        logger.info('Merging %s' % kwargs.get('heading'))
                        return section
                else:
                    logger.info('Skipping %s' % kwargs.get('heading'))
                    return None
            else:
                logger.info('Importing %s' % kwargs.get('heading'))

        section = self.make(Section, source_url=source_url, **kwargs)
        self.imported_section_ids.add(section.id)
        return section

    def get_tag(self, node):
        return etree.QName(node.tag).localname

    def get_text(self, node):
        paras = [node.text]
        paras += [
            etree.tostring(child, encoding='utf-8').decode('utf-8')
            for child in node.iterchildren()
            if self.get_tag(child) not in ('num', 'heading', 'subheading', 'from')
            ]
        return ''.join(filter(None, paras))

    def construct_heading(self, node):
        headings = {}
        for tag in ('num', 'heading', 'subheading'):
            if hasattr(node, tag):
                headings[tag] = getattr(node, tag).text
        return headings

    def construct_datetime(self, time):
        if not time:
            return (None, None)
        dt = dateutil.parse(time)
        return dt.date(), dt.time()

    def get_speaker(self, child):
        if hasattr(child, 'from'):
            display_name = child['from'].text
        else:
            display_name = None

        by_ref = child.get('by')
        speaker = None
        if by_ref:
            if not by_ref.startswith('#'):
                logger.warn(
                    "by attribute value doesn't begin with '#': %s" % by_ref)
            try:
                speaker = self.speakers[by_ref[1:]]
            except KeyError:
                logger.warn("ID %s in speech has no corresponding TLCPerson entry" % by_ref)

        return speaker, display_name

    def handle_tag(self, node, section):
        """If we need to do something out of the ordinary handling elements,
        subclass it here"""
        return False

    def visit(self, node, section):
        for child in node.iterchildren():
            tagname = self.get_tag(child)
            if tagname in ('num', 'heading', 'subheading'):
                # this will already have been extracted
                continue
            if tagname in (
                    'debateSection', 'administrationOfOath', 'rollCall',
                    'prayers', 'oralStatements', 'writtenStatements',
                    'personalStatements', 'ministerialStatements',
                    'resolutions', 'nationalInterest', 'declarationOfVote',
                    'communication', 'petitions', 'papers', 'noticesOfMotion',
                    'questions', 'address', 'proceduralMotions',
                    'pointOfOrder', 'adjournment',
                    ):
                headings = self.construct_heading(child)
                childSection = self.make_section(
                    parent=section,
                    start_date=self.start_date,
                    **headings
                )
                if childSection:
                    self.visit(child, childSection)
            elif tagname in ('speech', 'question', 'answer'):
                headings = self.construct_heading(child)
                text = self.get_text(child)
                start_date, start_time = self.construct_datetime(child.get('startTime'))
                end_date, end_time = self.construct_datetime(child.get('endTime'))
                speaker, display_name = self.get_speaker(child)
                self.make(
                    Speech,
                    section=section,
                    start_date=start_date or self.start_date,
                    start_time=start_time,
                    end_date=end_date,
                    end_time=end_time,
                    text=text,
                    speaker=speaker,
                    speaker_display=display_name,
                    type=tagname,
                    **headings
                    )
            elif tagname in ('scene', 'narrative', 'summary', 'other'):
                text = self.get_text(child)

                self.make(
                    Speech,
                    section=section,
                    start_date=self.start_date,
                    text=text,
                    type=tagname,
                    )
            else:
                success = self.handle_tag(child, section)
                if not success:
                    logger.error(
                        '%s unrecognised, "%s" - %s' %
                        (child.tag, child, self.get_text(child))
                        )



BLEACH_ALLOWED_TAGS = [
    'a', 'abbr', 'b', 'i', 'u', 'span', 'sub', 'sup', 'br',
    'p',
    'ol', 'ul', 'li',
    'table', 'caption', 'tr', 'th', 'td',
]

BLEACH_ALLOWED_ATTRIBUTES = {
    '*': ['id', 'title'],  # class, style
    'a': ['href'],
    'li': ['value'],
}
